# ------------------------------------------------------------------------------
#  File: tk_call_wrapper.py
#  Author: Jan Kukacka
#  Date: 11/2021
# ------------------------------------------------------------------------------
#  Call wrapper for custom error handling
#  based on https://stackoverflow.com/a/35390914/2042751
# ------------------------------------------------------------------------------

import tkinter as tk

class CallWrapper:
    def __init__(self, func, subst, widget):
        self.func = func
        self.subst = subst
        self.widget = widget

    def __call__(self, *args):
        try:
            if self.subst:
                args = self.subst(*args)
            return self.func(*args)
        except SystemExit as msg:
            raise SystemExit(msg)

        except tk.TclError as err:
            ## Raised when numbers from textboxes cannot be parsed
            pass

        except ValueError as err:
            ## Raised when colors from textboxes cannot be parsed
            pass



# ------------------------------------------------------------------------------
#  File: channels_list.py
#  Author: Jan Kukacka
#  Date: 11/2021
# ------------------------------------------------------------------------------
#  Viewer component for displaying the list of channels
# ------------------------------------------------------------------------------

import tkinter as tk
import tkinter.ttk as ttk

import os
import numpy as np
from pathlib import Path
from PIL import Image, ImageOps, ImageTk

try:
    from .scrolled_frame import ScrolledFrame
    from ..utils import event_handler
    from ..utils.tk_lazy_var import StringVar, BooleanVar
except ImportError:
    from tk_widgets.scrolled_frame import ScrolledFrame
    from utils import event_handler
    from utils.tk_lazy_var import StringVar, BooleanVar


def prepare_icon(filename, skin):
    path = Path(os.path.dirname(os.path.abspath(__file__)))
    icon = Image.open(str(path.parent/'resources'/filename))
    icon2 = icon.convert('RGB')
    icon2 = ImageOps.invert(icon2)
    icon2 = Image.fromarray(np.concatenate([np.array(icon2), np.array(icon)[...,-1,None]], axis=-1), mode='RGBA')
    bkg = Image.new(mode='RGBA', size=icon.size, color=skin.bg_color)
    icon = Image.alpha_composite(bkg, icon2).convert('RGB')
    icon = ImageTk.PhotoImage(icon)
    return icon

class ChannelsList():

    def __init__(self, parent_frame, skin, var_selected_channel):
        self.skin = skin

        self.var_selected_channel = var_selected_channel
        self.var_channels = []
        self.mainframe = ttk.LabelFrame(parent_frame, text='Channels', padding=5)
        self.mainframe.pack(side=tk.TOP, expand=False, fill=tk.BOTH, padx=5, pady=5)

        header_frame = tk.Frame(self.mainframe, bg=self.skin.bg_color)
        header_frame.pack(side=tk.TOP, expand=True, fill=tk.BOTH, padx=0, pady=0)

        icon_eye = prepare_icon('eye16.png', self.skin)
        label = tk.Label(header_frame, image=icon_eye, bg=self.skin.bg_color)
        label.image = icon_eye
        label.pack(side=tk.LEFT, padx=3)

        icon_color = prepare_icon('color16.png', self.skin)
        label = tk.Label(header_frame, image=icon_color, bg=self.skin.bg_color)
        label.image = icon_color
        label.pack(side=tk.LEFT, padx=0)

        tk.Label(header_frame, text='Channel name', bg=self.skin.bg_color, fg=self.skin.fg_color).pack(side=tk.LEFT, padx=3)

        self.wrap_frame = ScrolledFrame(self.mainframe, bg=self.skin.bg_color,
                                        use_ttk=True, scrollbars='vertical',
                                        borderwidth=0, relief=tk.FLAT,
                                        width=200, height=600)
        self.wrap_frame.pack(side=tk.TOP, expand=True, fill=tk.BOTH, padx=0, pady=0)
        self.wrap_frame.bind_scroll_wheel(self.mainframe)
        self.wrap_frame.bind_scroll_wheel(self.wrap_frame)

        self.item_frame = None

        footer_frame = tk.Frame(self.mainframe, bg=self.skin.bg_color)
        footer_frame.pack(side=tk.BOTTOM, expand=True, fill=tk.BOTH, padx=0, pady=0)
        self.btn_hide_all = ttk.Button(footer_frame, text='Hide all')
        self.btn_hide_all.pack(side=tk.LEFT, expand=False, padx=5, pady=5)
        self.btn_show_all = ttk.Button(footer_frame, text='Show all')
        self.btn_show_all.pack(side=tk.LEFT, expand=False, pady=5)


    def recreate_items(self, channel_props):
        if self.item_frame is not None:
            self.item_frame.destroy()
            self.item_frame = None

        self.items_frame = self.wrap_frame.display_widget(tk.Frame, fit_width=True)

        while len(self.var_channels) > 0:
            for var in self.var_channels[0].values():
                for trace_info in var.trace_vinfo():
                    var.trace_vdelete(*trace_info)
            self.var_channels.pop(0)

        for channel_prop in channel_props:
            self.var_channels.append({'color': StringVar(value=channel_prop['color']),
                                      'visible': BooleanVar(value=channel_prop['visible']),
                                      'name': StringVar(value='Channel name')})

        self.color_previews = []
        self.item_frames = []
        def add_channel_entry(channel_index, var_channel):
            bind_tag = f'ChannelSelect{channel_index}'
            frame = tk.Frame(self.items_frame, bg=self.skin.bg_color)
            frame.pack(side=tk.TOP, expand=True, fill=tk.BOTH, padx=0, pady=0)
            frame.bindtags((bind_tag,) + frame.bindtags())
            self.wrap_frame.bind_scroll_wheel(frame)
            self.item_frames.append(frame)

            frame.checkbox = ttk.Checkbutton(frame, text='', variable=var_channel['visible'],
                                             onvalue=True, offvalue=False)
            frame.checkbox.pack(side=tk.LEFT, expand=False, pady=2, padx=3)
            self.wrap_frame.bind_scroll_wheel(frame.checkbox)
            frame.checkbox.bindtags((bind_tag,) + frame.checkbox.bindtags())

            color_preview = tk.Frame(frame, width=16, height=16, bg=var_channel['color'].get(), highlightthickness=1, highlightbackground='#000000')
            color_preview.pack(side=tk.LEFT)
            color_preview.bindtags((bind_tag,) + color_preview.bindtags())

            self.wrap_frame.bind_scroll_wheel(color_preview)
            self.color_previews.append(color_preview)

            def on_colorchange(obj, var):
                obj.config(background=str(var.get()))
            handler = event_handler.TkVarEventHandler(on_colorchange, obj=color_preview, var=var_channel['color'])
            var_channel['color'].trace('w', handler)

            label = tk.Label(frame, textvariable=var_channel['name'], fg=self.skin.fg_color,
                             bg=self.skin.bg_color, anchor=tk.NW)
            label.pack(side=tk.LEFT, pady=2, padx=3)
            label.bindtags((bind_tag,) + label.bindtags())
            frame.label = label
            self.wrap_frame.bind_scroll_wheel(label)

            handler = event_handler.TkEventHandler(self.channel_onselect, var_channel_name=var_channel['name'])
            frame.bind_class(bind_tag, '<Button-1>', handler)
            handler = event_handler.TkEventHandler(self.init_rename_channel, channel_index=channel_index)
            frame.bind_class(bind_tag, '<Double-Button-1>', handler)

        for channel_index, var_channel in enumerate(self.var_channels):
            add_channel_entry(channel_index, var_channel)

    def update_vars(self, channel_props):
        for channel_property, var_channel in zip(channel_props, self.var_channels):
            for key in var_channel:
                var_channel[key].set(channel_property[key])


    def init_rename_channel(self, channel_index):
        frame = self.item_frames[channel_index]
        frame.label.pack_forget()
        entry = ttk.Entry(frame, width=25)
        entry.insert(0, self.var_channels[channel_index]['name'].get())
        entry.pack(side=tk.LEFT)
        entry.focus_set()
        entry.select_range(0, tk.END)
        handler = event_handler.TkEventHandler(self.end_rename_channel, frame=frame, entry=entry, channel_index=channel_index)
        entry.bind('<FocusOut>', handler)
        entry.bind('<Return>', handler)
        handler = event_handler.TkEventHandler(self.cancel_rename_channel, frame=frame, entry=entry)
        entry.bind('<Escape>', handler)

    def end_rename_channel(self, frame, entry, channel_index):
        value = entry.get()
        ## Check the value does not exist yet
        if value not in [vc['name'].get() for vc in self.var_channels]:
            old_value = self.var_channels[channel_index]['name'].get()
            self.var_channels[channel_index]['name'].set(value)
            ## Update also var_selected_channel if needed
            if self.var_selected_channel.get() == old_value:
                self.var_selected_channel.set(value)
        entry.destroy()
        frame.label.pack(side=tk.LEFT, pady=2, padx=3)

    def cancel_rename_channel(self, frame, entry):
        entry.destroy()
        frame.label.pack(side=tk.LEFT, pady=2, padx=3)

    @event_handler.requires('event')
    def channel_onselect(self, event, var_channel_name):
        event.widget.focus_set()
        self.var_selected_channel.set(var_channel_name.get())

    def highlight_item(self, item_index):
        for frame in self.item_frames:
            frame.config(bg=self.skin.bg_color)
            frame.checkbox.configure(style='TCheckbutton')
            frame.label.config(bg=self.skin.bg_color)
        self.item_frames[item_index].config(bg=self.skin.bg_highlight_color)
        self.item_frames[item_index].checkbox.configure(style='Highlight.TCheckbutton')
        self.item_frames[item_index].label.config(bg=self.skin.bg_highlight_color)



# ------------------------------------------------------------------------------
#  File: observable.py
#  Author: Remi Emonet
#  Date: 8/2019
#  Source: https://github.com/twitwi/vuejs-python/tree/master/vuejspython/observablecollections
#  License: MIT
# ------------------------------------------------------------------------------

import types
from .event import Event

class Observable:
	def __init__(self):
		self.observers = set()

	def attach(self, observer):
		if (observer not in self.observers):
			self.observers.add(observer)

	def detach(self, observer):
		if (observer in self.observers):
			self.observers.remove(observer)

	def raiseEvent(self, name, **kwargs):
		event = Event(name, self)

		for key, value in kwargs.items():
			setattr(event, key, value)

		for eventObserver in self.observers:
			eventObserver(event)



# ------------------------------------------------------------------------------
#  File: pipeline.py
#  Author: Jan Kukacka
#  Date: 2/2022
# ------------------------------------------------------------------------------
#  Filtering pipeline
# ------------------------------------------------------------------------------


try:
    from . import filter_factory
except ImportError:
    from filters import filter_factory

class Pipeline(object):
    '''
    Pipeline contains a sequence of filters to pass an image through
    '''

    def __init__(self, filters):
        self.filters = filters

    def __call__(self, img):
        for filter in self.filters:
            img = filter(img)
        return img

    def serialize(self):
        return {'filters': [filter.serialize() for filter in self.filters]}

    @staticmethod
    def call(serialization, img):
        for filter in serialization['filters']:
            params = filter['params']
            if params['active']:
                T_filter = filter_factory.get_filter_by_name(filter['name'])
                img = T_filter.call(img, **params)
        return img

    @staticmethod
    def deserialize(serialization):
        filters = []
        for filter in serialization['filters']:
            T_filter = filter_factory.get_filter_by_name(filter['name'])
            filter = T_filter.deserialize(filter['params'])
            filters.append(filter)
        return Pipeline(filters)


    def update(self, serialization):
        '''
        Update filters in the pipeline and invalidates cache of existing filters
        if a change is detected in an earlier stage

        # Arguments:
            - serialization: Dict with serialized pipeline.

        # Returns:
            - change_detected: Bool. True if Pipeline was updated, False if
                serialization was matching the current pipeline.
        '''
        change_detected = False
        for i, filter in enumerate(serialization['filters']):
            if i >= len(self.filters):
                T_filter = filter_factory.get_filter_by_name(filter['name'])
                new_filter = T_filter.deserialize(filter['params'])
                self.filters.append(new_filter)
                change_detected = True
            if self.filters[i].serialize() != filter:
                T_filter = filter_factory.get_filter_by_name(filter['name'])
                new_filter = T_filter.deserialize(filter['params'])
                self.filters[i] = new_filter
                change_detected = True
            if change_detected:
                self.filters[i].cache = None
        if len(self.filters) > len(serialization['filters']):
            self.filters = self.filters[:len(serialization['filters'])]
            change_detected = True

        return change_detected



# ------------------------------------------------------------------------------
#  File: event_handler.py
#  Author: Jan Kukacka
#  Date: 2/2022
# ------------------------------------------------------------------------------
#  Classes for event handling and monitoring event flow in the app
# ------------------------------------------------------------------------------

from functools import wraps
from time import time

show_events = False
event_depth = 0

class EventHandler:
    '''
    Base class for event handlers
    '''
    def __init__(self, func, name_=None, **kwargs):
        self.func = func
        self.func_kwargs = kwargs
        self.name = name_

    def __call__(self, *args, event_details=None, **kwargs):
        global show_events, event_depth
        if not show_events:
            self.func(*args, **kwargs)
        else:
            event_depth += 1
            try:
                print(f'{"│  "*(event_depth-2)}{"◇" if event_depth==1 else "├─"}', f'Event ({self.name}): handler = {self.func.__name__} (Details: {event_details})')

                t0 = time()
                self.func(*args, **kwargs)
                t1 = time()
                print(f'{"│  "*(event_depth-2)}{"└◈" if event_depth==1 else "├──┘"}', f' [{t1-t0:.5f} s]')
            finally:
                event_depth -= 1



class TkVarEventHandler(EventHandler):
    def __init__(self, func, **kwargs):
        super().__init__(func, **kwargs)

    def __call__(self, varname, varindex, action):
        args = []
        if hasattr(self.func, 'requires'):
            if 'varname' in self.func.requires:
                args.append(varname)
            if 'varindex' in self.func.requires:
                args.append(varindex)
            if 'action' in self.func.requires:
                args.append(action)
        super().__call__(event_details=f'trigger var = {varname}', *args, **self.func_kwargs)


class ObservableEventHandler(EventHandler):
    def __init__(self, func, **kwargs):
        super().__init__(func, **kwargs)

    def __call__(self, event):
        event_details = f'source = {event.source}, action = {event.action}'
        if event.action == 'propertyChanged' and hasattr(event, 'propertyName'):
            event_details += f'[{event.propertyName}]'
        try:
            if 'event' in self.func.requires:
                super().__call__(event, event_details=event_details, **self.func_kwargs)
            else:
                super().__call__(event_details=event_details,**self.func_kwargs)
        except Exception as e:
            super().__call__(event_details=event_details,**self.func_kwargs)


class TkCommandEventHandler(EventHandler):
    def __init__(self, func, **kwargs):
        super().__init__(func, **kwargs)

    def __call__(self):
        super().__call__(**self.func_kwargs)

class TkEventHandler(EventHandler):
    def __init__(self, func, **kwargs):
        super().__init__(func, **kwargs)

    def __call__(self, event):
        event_details = f'source = {event.widget}, action = {event.type}'
        try:
            if 'event' in self.func.requires:
                super().__call__(event, event_details=event_details, **self.func_kwargs)
            else:
                super().__call__(event_details=event_details, **self.func_kwargs)
        except Exception as e:
            super().__call__(event_details=event_details, **self.func_kwargs)


def requires(argument_name):
    def decorator(func):
        if not hasattr(func, 'requires'):
            func.requires = {}
        func.requires[argument_name] = True
        return func
    return decorator



# ------------------------------------------------------------------------------
#  File: anisotropic_denoising.py
#  Author: Jan Kukacka
#  Date: 2/2022
# ------------------------------------------------------------------------------
#  Implementation of anisotropic denoising
#  Based on code by Alistair Muldal (c) 2012
#      - source: https://pastebin.com/sBsPX4Y7
# ------------------------------------------------------------------------------

import numpy as np
from cv2.ximgproc import anisotropicDiffusion

try:
    from . import filter
except ImportError:
    from filters import filter

class AnisotropicDenoising(filter.Filter):
    '''

    '''

    name = 'anisotropic_denoising'

    def __init__(self, step_size=.15, sensitivity=.1, n_iter=10):
        '''
        # Arguments:
            - step_size:
            - sensitivity:
            - n_iter:
        '''
        super().__init__()
        self.step_size = step_size
        self.sensitivity = sensitivity
        self.n_iter = n_iter

    def __call__(self, img):
        '''
        # Arguments:
            - img: array of data to be normalized

        # Returns:
            - normalized array.
        '''
        result = super().__call__(img)
        if result is not None:
            return result
        else:
            self.cache = self.call(img, self.step_size, self.sensitivity, self.n_iter)
            return self.cache


    @staticmethod
    def call(img, step_size, sensitivity, n_iter, **kwargs):
        img_min, img_max = img.min(), img.max()

        norm_img = anisodiff(img, step_size, sensitivity, n_iter)

        norm_min, norm_max = norm_img.min(), norm_img.max()
        scale = (img_max-img_min) / (norm_max-norm_min)
        return img_min + (norm_img-norm_min) * scale


    def serialize(self):
        base_dict = super().serialize()
        base_dict['params']['step_size'] = self.step_size
        base_dict['params']['sensitivity'] = self.sensitivity
        base_dict['params']['n_iter'] = self.n_iter
        return base_dict

    @staticmethod
    def deserialize(serialization):
        step_size = serialization['step_size']
        sensitivity = serialization['sensitivity']
        n_iter = serialization['n_iter']
        obj = AnisotropicDenoising(step_size, sensitivity, n_iter)
        obj._deserialize_parent(serialization)
        return obj

# ------------------------------------------------------------------------------
#  Based on code by Alistair Muldal (c) 2012
#      - source: https://pastebin.com/sBsPX4Y7

# Original work: Copyright (c) 1995-2012 Peter Kovesi pk@peterkovesi.com
# Modified work: Copyright (c) 2012 Alistair Muldal
# Modified work: Copyright (c) 2022 Jan Kukacka
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# The software is provided "as is", without warranty of any kind, express or
# implied, including but not limited to the warranties of merchantability,
# fitness for a particular purpose and noninfringement. In no event shall the
# authors or copyright holders be liable for any claim, damages or other
# liability, whether in an action of contract, tort or otherwise, arising from,
# out of or in connection with the software or the use or other dealings in the
# software.

from scipy.ndimage import gaussian_filter

def anisodiff(img, gamma=0.1, kappa=50, niter=1, sigma=0, option=1):
    """
    Anisotropic diffusion.

    Usage:
    imgout = anisodiff(im, niter, kappa, gamma, option)

    Arguments:
            img    - input image
            gamma  - max value of .25 for stability
            kappa  - conduction coefficient 20-100 ?
            niter  - number of iterations
            step   - tuple, the distance between adjacent pixels in (y,x)
            option - 1 Perona Malik diffusion equation No 1
                     2 Perona Malik diffusion equation No 2

    Returns:
            imgout   - diffused image.

    kappa controls conduction as a function of gradient.  If kappa is low
    small intensity gradients are able to block conduction and hence diffusion
    across step edges.  A large value reduces the influence of intensity
    gradients on conduction.

    gamma controls speed of diffusion (you usually want it at a maximum of
    0.25)

    Diffusion equation 1 favours high contrast edges over low contrast ones.
    Diffusion equation 2 favours wide regions over smaller ones.
    """

    # initialize output array
    imgout = img.copy()

    # initialize some internal variables
    deltaS = np.zeros_like(imgout)
    deltaE = deltaS.copy()
    NS = deltaS.copy()
    EW = deltaS.copy()
    gS = np.ones_like(imgout)
    gE = gS.copy()

    for ii in np.arange(1,niter):

        # calculate the diffs
        deltaS[:-1,: ] = np.diff(imgout,axis=0)
        deltaE[: ,:-1] = np.diff(imgout,axis=1)

        if sigma > 0:
            deltaSf=gaussian_filter(deltaS,sigma)
            deltaEf=gaussian_filter(deltaE,sigma)
        else:
            deltaSf=deltaS
            deltaEf=deltaE

        # conduction gradients (only need to compute one per dim!)
        if option == 1:
            gS = np.exp(-(deltaSf/kappa)**2.)
            gE = np.exp(-(deltaEf/kappa)**2.)
        elif option == 2:
            gS = 1./(1.+(deltaSf/kappa)**2.)
            gE = 1./(1.+(deltaEf/kappa)**2.)

        # update matrices
        E = gE*deltaE
        S = gS*deltaS

        # subtract a copy that has been shifted 'North/West' by one
        # pixel. don't as questions. just do it. trust me.
        NS[:] = S
        EW[:] = E
        NS[1:,:] -= S[:-1,:]
        EW[:,1:] -= E[:,:-1]

        # update the image
        imgout += gamma*(NS+EW)

    return imgout

# ------------------------------------------------------------------------------



# ------------------------------------------------------------------------------
#  File: gaussian_blur.py
#  Author: Jan Kukacka
#  Date: 2/2022
# ------------------------------------------------------------------------------
#  Implementation of Gaussian blur
# ------------------------------------------------------------------------------

import numpy as np
from skimage.filters import gaussian

try:
    from . import filter
except ImportError:
    from filters import filter

class GaussianBlur(filter.Filter):
    '''
    Performs gaussian blur
    '''

    name = 'gaussian_blur'

    def __init__(self, sigma=1):
        '''
        # Arguments:
            - sigma: Std. deviation of the gaussian kernel
        '''
        super().__init__()
        self.sigma = sigma

    def __call__(self, img):
        '''
        # Arguments:
            - img: array of data to be normalized

        # Returns:
            - normalized array.
        '''
        result = super().__call__(img)
        if result is not None:
            return result
        else:
            self.cache = self.call(img, self.sigma)
            return self.cache


    @staticmethod
    def call(img, sigma, **kwargs):
        img_min, img_max = img.min(), img.max()

        norm_img = gaussian(img, sigma, preserve_range=True)

        norm_min, norm_max = norm_img.min(), norm_img.max()
        scale = (img_max-img_min) / (norm_max-norm_min)
        return img_min + (norm_img-norm_min) * scale
        return norm_img

    def serialize(self):
        base_dict = super().serialize()
        base_dict['params']['sigma'] = self.sigma
        return base_dict

    @staticmethod
    def deserialize(serialization):
        sigma = serialization['sigma']
        obj = GaussianBlur(sigma)
        obj._deserialize_parent(serialization)
        return obj



# ------------------------------------------------------------------------------
#  File: windnd.py
#  Author: Jan Kukacka
#  Date: 3/2021
# ------------------------------------------------------------------------------
#  Windows File Drag'n'Drop plugin
#  Code from: https://github.com/cilame/windnd/blob/master/windnd/windnd.py
# ------------------------------------------------------------------------------



def hook_dropfiles(tkwindow_or_winfoid,func,force_unicode=False):
    """
    # this func to deal drag icon & drop to load in windows
    *args:
        hwnd
    **kw:
        func = _func
        # default func just use path list print each one.
        # default each full_path_file_name type: bytes; in py3
        # default each full_path_file_name type: str  ; in py2
        #===================
        def _func(ls):
            for i in ls:
                print(i)
        #===================
    test evironment:
        py2, py3
        work on win7 32bit & 64bit.

    if you use tk, you can hook like this:
    =================================================
    > import windnd
    >
    > def my_func(ls):
          for idx,i in enumerate(ls):
              print(idx,i)
    >
    > import tkinter
    > tk = tkinter.Tk()
    > hwnd = tk.winfo_id()
    >
    > # you don't have to write "hwnd = tk.winfo_id()" in tkinter
    > # because you can put "tk" in this function like:
    > # "windnd.hook_dropfiles(tk,func = my_func)"
    > # the reason for this is to expand interface
    >
    > windnd.hook_dropfiles(hwnd,func = my_func)
    >
    > tk.mainloop()
    =================================================
    """

    # this place just for expand interface
    # because may anther window tools need use hwnd to hook
    # If you want to process .lnk , you can look at examples from the source code

    import platform
    import ctypes
    from ctypes.wintypes import DWORD

    hwnd = tkwindow_or_winfoid.winfo_id()\
           if getattr(tkwindow_or_winfoid, "winfo_id", None)\
           else tkwindow_or_winfoid

    if platform.architecture()[0] == "32bit":
        GetWindowLong = ctypes.windll.user32.GetWindowLongW
        SetWindowLong = ctypes.windll.user32.SetWindowLongW
        argtype = DWORD
    elif platform.architecture()[0] == "64bit":
        GetWindowLong = ctypes.windll.user32.GetWindowLongPtrA
        SetWindowLong = ctypes.windll.user32.SetWindowLongPtrA
        argtype = ctypes.c_uint64

    prototype = ctypes.WINFUNCTYPE(argtype,argtype,argtype,argtype,argtype)
    WM_DROPFILES = 0x233
    GWL_WNDPROC = -4
    create_buffer = ctypes.create_unicode_buffer if force_unicode else ctypes.c_buffer
    func_DragQueryFile = ctypes.windll.shell32.DragQueryFileW if force_unicode else ctypes.windll.shell32.DragQueryFile

    def py_drop_func(hwnd,msg,wp,lp):
        global files
        if msg == WM_DROPFILES:
            count = func_DragQueryFile(argtype(wp),-1,None,None)
            szFile = create_buffer(260)
            files = []
            for i in range(count):
                func_DragQueryFile(argtype(wp),i,szFile,ctypes.sizeof(szFile))
                dropname = szFile.value
                files.append(dropname)
            func(files)
            ctypes.windll.shell32.DragFinish(argtype(wp))
        return ctypes.windll.user32.CallWindowProcW(*map(argtype,(globals()[old],hwnd,msg,wp,lp)))

    # for limit hook number, protect computer.
    limit_num = 200
    for i in range(limit_num):
        if i+1 == limit_num:
            raise "over hook limit number 200, for protect computer."
        if "old_wndproc_%d" % i not in globals():
            old, new = "old_wndproc_%d"%i, "new_wndproc_%d"%i
            break

    globals()[old] = None
    globals()[new] = prototype(py_drop_func)

    ctypes.windll.shell32.DragAcceptFiles(hwnd,True)
    globals()[old] = GetWindowLong(hwnd,GWL_WNDPROC)
    SetWindowLong(hwnd,GWL_WNDPROC,globals()[new])


# if __name__ == '__main__':
#     def myfunc(ls):
#         def _local_lnk(link):
#             ''' 处理 link 指向的问题。'''
#             import platform
#             if platform.python_version().startswith('3') and type(link) is bytes:
#                 try:
#                     _link = link.decode()
#                 except:
#                     _link = link.decode('gbk')
#             else:
#                 _link = link
#             try:
#                 import sys, win32com.client
#                 shell = win32com.client.Dispatch("WScript.Shell")
#                 shortcut = shell.CreateShortCut(_link)
#                 return shortcut.Targetpath if shortcut.Targetpath.strip() else _link
#             except:
#                 return _link
#         for i in ls:
#             print('deal link:',_local_lnk(i))
#
#     def test():
#         '''
#         将 windows 桌面的图标拖拽到被挂钩的窗口内
#         对加载到的图标路径进行对应处理的函数。
#         '''
#         import tkinter
#         tk = tkinter.Tk()
#         hook_dropfiles(tk, myfunc)
#         tk.mainloop()
#
#     test()



# ------------------------------------------------------------------------------
#  File: observablelist.py
#  Author: Remi Emonet
#  Date: 8/2019
#  Source: https://github.com/twitwi/vuejs-python/tree/master/vuejspython/observablecollections
#  License: MIT
# ------------------------------------------------------------------------------

from .observable import Observable

class ObservableList(list, Observable):

	def __init__(self, *args, **kwargs):
		Observable.__init__(self)
		list.__init__(self, *args, **kwargs)

	def __iadd__(self, other):
		self.extend(other)
		return self

	def append(self, value):
		index = len(self)
		list.append(self, value)
		self.raiseEvent('itemsAdded', index=index, items=[value])

	def insert(self, index, value):
		list.insert(self, index, value)

		if index > len(self) - 1:
			index = len(self) - 1
		self.raiseEvent('itemsAdded', index=index, items=[value])

	def __setitem__(self, key, value):
		if (type(key) is slice):
			index = key.start
		else:
			index = key

		oldItem = list.__getitem__(self, key)

		new_is_different = True
		try:
			if oldItem == value:
				new_is_different = False
		except:
			print('x')
			pass
		if new_is_different:
			list.__setitem__(self, key, value)
			self.raiseEvent('itemsUpdated', index=index, oldItems=[oldItem], newItems=[value])

	def __delitem__(self, i):
		if isinstance(i, slice):
			index = i.start
		else:
			index = i

		item = list.__getitem__(self, i)
		list.__delitem__(self, i)
		self.raiseEvent('itemsRemoved', items=[item], index=index)

	def extend(self, seq):
		insertIndex = len(self)
		list.extend(self, seq)
		self.raiseEvent('itemsAdded', items=seq, index=insertIndex)

	def pop(self, index=-1):
		removedIndex = index

		if (index == -1):
			removedIndex = len(self) - 1

		value = list.pop(self, index)

		self.raiseEvent('itemsRemoved', items=[value], index=removedIndex)
		return value

	def remove(self, obj):
		index = self.index(obj)
		list.remove(self, obj)
		self.raiseEvent('itemsRemoved', items=[obj], index=index)

	def sort(self, **kwargs):
		list.sort(self, **kwargs)
		self.raiseEvent('sorted')

	def reverse(self):
		list.reverse(self)
		self.raiseEvent('reversed')



# ------------------------------------------------------------------------------
#  File: presenter.py
#  Author: Jan Kukacka
#  Date: 3/2021
# ------------------------------------------------------------------------------
#  Presenter component
# ------------------------------------------------------------------------------

import numpy as np
import happy as hp
from matplotlib.colors import is_color_like, to_hex

try:
    from . import view as view_
    from .utils.windnd import hook_dropfiles
    from .utils import event_handler
except ImportError:
    import view as view_
    from utils import event_handler
    from utils.windnd import hook_dropfiles

class Presenter(object):
    '''
    '''

    def __init__(self, view, model):
        self.view = view
        self.model = model

        ## Define view callbacks
        ## -- Update color previews
        handler = event_handler.TkVarEventHandler(self.update_widget_bkg, obj=self.view.color_preview, var=self.view.var_channel['color'])
        self.view.var_channel['color'].trace('w', handler)

        ## -- Color picker clicks
        handler = event_handler.TkCommandEventHandler(self.pick_color, var=self.view.var_channel['color'])
        self.view.color_picker.config(command=handler)

        ## -- Update colormap preview
        self.view.var_channel['color'].trace('w', event_handler.TkVarEventHandler(self.color_onchage))

        ## -- channel property changes
        self.view.var_selected_channel.trace('w', event_handler.TkVarEventHandler(self.channel_onchange))
        for key,var in self.view.var_channel.items():
            var.trace('w', event_handler.TkVarEventHandler(self.channel_var_onchange, key=key, var=var))

        ## -- channel control buttons
        self.view.channels_panel.btn_hide_all.config(command=event_handler.TkCommandEventHandler(self.hide_all))
        self.view.channels_panel.btn_show_all.config(command=event_handler.TkCommandEventHandler(self.show_all))

        ## -- config saving and loading
        self.view.btn_save.config(command=event_handler.TkCommandEventHandler(self.save_model))
        self.view.btn_load.config(command=event_handler.TkCommandEventHandler(self.load_model))
        self.view.btn_save_render.config(command=event_handler.TkCommandEventHandler(self.save_render))

        ## -- config copying and pasting
        def paste_params(obj):
            obj.model.paste_params(obj.view.get_active_channel())
        self.view.btn_paste.config(command=event_handler.TkCommandEventHandler(paste_params, obj=self))
        def copy_params(obj):
            obj.model.copy_params(obj.view.get_active_channel())
        self.view.btn_copy.config(command=event_handler.TkCommandEventHandler(copy_params, obj=self))

        ## -- clipboard
        self.view.bind('<Control-c>', event_handler.TkEventHandler(self.render_to_clipboard))
        self.view.bind('<Control-s>', event_handler.TkEventHandler(self.save_render))
        self.view.bind('<Control-o>', event_handler.TkEventHandler(self.load_image))
        self.view.bind('<Control-t>', event_handler.TkEventHandler(self.model.transpose_image))

        ## -- bind menu commands
        self.view.menu['file']['obj'].entryconfig(self.view.menu['file']['load_image'], command=event_handler.TkCommandEventHandler(self.load_image))
        self.view.menu['file']['obj'].entryconfig(self.view.menu['file']['save_render'], command=event_handler.TkCommandEventHandler(self.save_render))
        self.view.menu['file']['obj'].entryconfig(self.view.menu['file']['save_config'], command=event_handler.TkCommandEventHandler(self.save_model))
        self.view.menu['file']['obj'].entryconfig(self.view.menu['file']['load_config'], command=event_handler.TkCommandEventHandler(self.load_model))
        self.view.menu['image']['obj'].entryconfig(self.view.menu['image']['transpose'], command=event_handler.TkCommandEventHandler(self.model.transpose_image))
        self.view.menu['image']['obj'].entryconfig(self.view.menu['image']['autocolor'], command=event_handler.TkCommandEventHandler(self.model.autocolor))

        ## -- adding filters
        for key, val in self.view.menu_add_filter.items():
            if key == 'obj': continue
            def add_filter(filter):
                self.model.add_filter(self.view.get_active_channel(), filter_name=filter)
            self.view.menu_add_filter['obj'].entryconfig(
                self.view.menu_add_filter[key],
                command=event_handler.TkCommandEventHandler(add_filter, filter=key))

        ## -- on closing
        # self.protocol("WM_DELETE_WINDOW", self.on_closing)

        ## Define model callbacks
        self.model.attach(event_handler.ObservableEventHandler(self.model_onchange))

        ## Hook on model rendering updates
        self.check_model_for_render()

        ## Hook drag'n'drop files
        hook_dropfiles(self.view, self.drag_file)

    def register_channel_panel_handlers(self):
        ## Register handlers on variables of channels panel
        for i,var_channel in enumerate(self.view.channels_panel.var_channels):
            for key,var in var_channel.items():
                # if key != 'name':  ## Name has a seperate handling mechanism
                var.trace('w', event_handler.TkVarEventHandler(self.channel_var_onchange, var=var,key=key,cindex=i))

        for i, color_preview in enumerate(self.view.channels_panel.color_previews):
            color_preview.bind('<Double-Button-1>', event_handler.TkEventHandler(self.pick_color, var=self.view.channels_panel.var_channels[i]['color']))


    def check_model_for_render(self):
        self.model.check_for_render()
        self.view.after(50, self.check_model_for_render)

    def check_model_for_io(self):
        self.model.check_for_io()
        if self.model.n_io_pending > 0:
            self.view.after(50, self.check_model_for_io)
        else:
            self.view.loader.hide()

    @event_handler.requires('event')
    def model_onchange(self, event):
        '''
        Propagates changes in the model to the view.
        '''
        if event.action == 'ioTask':
            self.view.loader.show()
            self.check_model_for_io()

        if event.action == 'propertyChanged':
            if event.propertyName == 'channel_props':
                if hasattr(event, 'child'):
                    if event.child.action == 'itemsAdded' or event.child.action == 'itemsRemoved':
                        self.view.update_channels(self.model.channel_props)
                        self.register_channel_panel_handlers()

                self.view.channels_panel.update_vars(self.model.channel_props)
                active_channel = self.view.get_active_channel()
                if active_channel < len(self.model.channel_props):
                    self.channel_onchange()
            if event.propertyName == 'render':
                if self.model.render is not None:
                    self.view.show_image(self.model.render)

                if self.model.response_images is not None:
                    channel_index = self.view.get_active_channel()
                    self.view.show_response(self.model.response_images[channel_index])


    def save_model(self):
        model_dict = self.model.save()
        filename = self.view.asksaveasfilename(title='Save config as...', filetypes=[('JSON files', '.json')], initialfile='config.json')
        try:
            hp.io.save(filename, model_dict, overwrite=True)
        except Exception as e:
            print('Error saving model as', filename)
            print(e)

    def load_model(self):
        filename = self.view.askopenfilename(title='Load config file...', filetypes=[('JSON files', '.json')])
        try:
            self.model.load(hp.io.load(filename))
        except Exception as e:
            print(e)

    def load_image(self, *args):
        filename = self.view.askopenfilename(title='Load image...', filetypes=[('NIFTI images', '.nii.gz'), ('Numpy arrays', '.npy'), ('MATLAB', '.mat'), ('NRRD images', '.nrrd')])
        try:
            self.model.filename = filename
        except Exception as e:
            print(e)

    def drag_file(self, files):
        if len(files) > 0:
            ## Model filename update has to be registered using "after",
            ## otherwise the program crashes with:
            ## Fatal Python error: PyEval_RestoreThread: NULL tstate
            filename = files[0].decode()

            ## If file is json, try loading it as a model
            if filename.lower().endswith('.json'):
                def update_model():
                    try:
                        self.model.load(hp.io.load(filename))
                    except Exception as e:
                        print(e)
                self.view.after(10, update_model)
            ## Else try loading it as an image
            else:
                def update_filename():
                    self.model.filename = filename
                self.view.after(10, update_filename)

    def save_render(self, *args):
        filename = self.view.asksaveasfilename(title='Save render as...', filetypes=[('Numpy files', '.npy'), ('PNG', '.png')], initialfile='render.npy')
        try:
            hp.io.save(filename, self.model.render, overwrite=True)
        except Exception as e:
            print('Error saving render as', filename)
            print(e)

    def channel_onchange(self):
        channel_index = self.view.get_active_channel()
        if channel_index < 0 or channel_index >= len(self.model.channel_props):
            return

        channel_property = self.model.channel_props[channel_index]

        for key, val in channel_property.items():
            if key in self.view.var_channel:
                self.view.var_channel[key].set(val)
            if key in self.view.property_frames:
                view_.set_state(self.view.property_frames[key], 'normal' if val else 'disable')

        self.view.channels_panel.highlight_item(channel_index)
        self.view.pipelines_panel.on_channel_selected_change(channel_index)
        if self.model.response_images is not None:
            self.view.show_response(self.model.response_images[channel_index])


    def color_onchage(self):
        channel_index = self.view.get_active_channel()
        var = self.view.var_channel['color']
        color = var.get()
        if is_color_like(color) and channel_index < len(self.model.channel_props):
            self.model.channel_props[channel_index]['color'] = str(to_hex(color))

    def channel_var_onchange(self, var, key, cindex=None):
        if cindex is None:
            cindex = self.view.get_active_channel()
        self.model.channel_props[cindex][key] = var.get()


    @staticmethod
    def update_widget_bkg(obj,var):
        color = var.get()
        try:
            obj.config(background=str(to_hex(color)))
        except:
            pass

    def pick_color(self, var):
        cindex = self.view.get_active_channel()
        color = var.get()
        try:
            response = self.view.askcolor(initialcolor=str(to_hex(color)))
            if response is not None:
                var.set(response)
        except:
            pass

    def hide_all(self, *_):
        cindex = self.view.get_active_channel()
        for i in range(len(self.model.channel_props)):
            if i != cindex:
                self.model.channel_props[i]['visible'] = False

    def show_all(self, *_):
        for i in range(len(self.model.channel_props)):
            self.model.channel_props[i]['visible'] = True

    def render_to_clipboard(self, _):
        '''
        Code from https://stackoverflow.com/a/62007792/2042751
        '''
        from io import BytesIO
        import win32clipboard
        from PIL import Image

        def send_to_clipboard(clip_type, data):
            win32clipboard.OpenClipboard()
            win32clipboard.EmptyClipboard()
            win32clipboard.SetClipboardData(clip_type, data)
            win32clipboard.CloseClipboard()

        # image = self.model.render
        with BytesIO() as stream:
            self.model.render.convert("RGB").save(stream, "BMP")
            data = stream.getvalue()[14:]
            send_to_clipboard(win32clipboard.CF_DIB, data)

    def mainloop(self):
        self.view.mainloop()



# ------------------------------------------------------------------------------
#  File: scrolled_frame.py
#  Author: Benjamin Johnson
#  Date: 9/2019
#  Source: https://github.com/bmjcode/tkScrolledFrame
#  License: MIT
# ------------------------------------------------------------------------------
#  Modified by Jan Kukacka, 11/2021
# ------------------------------------------------------------------------------
"""Implementation of the scrollable frame widget."""

import sys

try:
    # Python 3
    import tkinter as tk
except (ImportError):
    # Python 2
    import Tkinter as tk

try:
    try:
        # Python 3
        import tkinter.ttk as ttk
    except (ImportError):
        # Python 2
        import ttk
except (ImportError):
    # Can't provide ttk's Scrollbar
    pass


# __all__ = ["ScrolledFrame"]


class ScrolledFrame(tk.Frame):
    """Scrollable Frame widget.

    Use display_widget() to set the interior widget. For example,
    to display a Label with the text "Hello, world!", you can say:

        sf = ScrolledFrame(self)
        sf.pack()
        sf.display_widget(Label, text="Hello, world!")

    The constructor accepts the usual Tkinter keyword arguments, plus
    a handful of its own:

      scrollbars (str; default: "both")
        Which scrollbars to provide.
        Must be one of "vertical", "horizontal," "both", or "neither".

      use_ttk (bool; default: False)
        Whether to use ttk widgets if available.
        The default is to use standard Tk widgets. This setting has
        no effect if ttk is not available on your system.
    """

    def __init__(self, master=None, **kw):
        """Return a new scrollable frame widget."""

        tk.Frame.__init__(self, master)

        # Hold these names for the interior widget
        self._interior = None
        self._interior_id = None

        # Whether to fit the interior widget's width to the canvas
        self._fit_width = False

        # Which scrollbars to provide
        if "scrollbars" in kw:
            scrollbars = kw["scrollbars"]
            del kw["scrollbars"]

            if not scrollbars:
                scrollbars = self._DEFAULT_SCROLLBARS
            elif not scrollbars in self._VALID_SCROLLBARS:
                raise ValueError("scrollbars parameter must be one of "
                                 "'vertical', 'horizontal', 'both', or "
                                 "'neither'")
        else:
            scrollbars = self._DEFAULT_SCROLLBARS

        # Whether to use ttk widgets if available
        if "use_ttk" in kw:
            if ttk and kw["use_ttk"]:
                Scrollbar = ttk.Scrollbar
            else:
                Scrollbar = tk.Scrollbar
            del kw["use_ttk"]
        else:
            Scrollbar = tk.Scrollbar

        # Default to a 1px sunken border
        if not "borderwidth" in kw:
            kw["borderwidth"] = 1
        if not "relief" in kw:
            kw["relief"] = "sunken"

        # Set up the grid
        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)

        # Canvas to hold the interior widget
        c = self._canvas = tk.Canvas(self,
                                     borderwidth=0,
                                     highlightthickness=0,
                                     takefocus=0)

        # Enable scrolling when the canvas has the focus
        self.bind_arrow_keys(c)
        self.bind_scroll_wheel(c)

        # Call _resize_interior() when the canvas widget is updated
        c.bind("<Configure>", self._resize_interior)

        # Scrollbars
        xs = self._x_scrollbar = Scrollbar(self,
                                           orient="horizontal",
                                           command=c.xview)
        ys = self._y_scrollbar = Scrollbar(self,
                                           orient="vertical",
                                           command=c.yview)
        c.configure(xscrollcommand=xs.set, yscrollcommand=ys.set)

        # Lay out our widgets
        c.grid(row=0, column=0, sticky="nsew")
        if scrollbars == "vertical" or scrollbars == "both":
            ys.grid(row=0, column=1, sticky="ns")
        if scrollbars == "horizontal" or scrollbars == "both":
            xs.grid(row=1, column=0, sticky="we")

        # Forward these to the canvas widget
        self.bind = c.bind
        self.focus_set = c.focus_set
        self.unbind = c.unbind
        self.xview = c.xview
        self.xview_moveto = c.xview_moveto
        self.yview = c.yview
        self.yview_moveto = c.yview_moveto

        # Process our remaining configuration options
        self.configure(**kw)

    def __setitem__(self, key, value):
        """Configure resources of a widget."""

        if key in self._CANVAS_KEYS:
            # Forward these to the canvas widget
            self._canvas.configure(**{key: value})

        else:
            # Handle everything else normally
            tk.Frame.configure(self, **{key: value})

    # ------------------------------------------------------------------------

    def bind_arrow_keys(self, widget):
        """Bind the specified widget's arrow key events to the canvas."""

        widget.bind("<Up>",
                    lambda event: self._canvas.yview_scroll(-1, "units"))

        widget.bind("<Down>",
                    lambda event: self._canvas.yview_scroll(1, "units"))

        widget.bind("<Left>",
                    lambda event: self._canvas.xview_scroll(-1, "units"))

        widget.bind("<Right>",
                    lambda event: self._canvas.xview_scroll(1, "units"))

    def bind_scroll_wheel(self, widget):
        """Bind the specified widget's mouse scroll event to the canvas."""

        widget.bind("<MouseWheel>", self._scroll_canvas)
        widget.bind("<Button-4>", self._scroll_canvas)
        widget.bind("<Button-5>", self._scroll_canvas)

    def cget(self, key):
        """Return the resource value for a KEY given as string."""

        if key in self._CANVAS_KEYS:
            return self._canvas.cget(key)

        else:
            return tk.Frame.cget(self, key)

    # Also override this alias for cget()
    __getitem__ = cget

    def configure(self, cnf=None, **kw):
        """Configure resources of a widget."""

        # This is overridden so we can use our custom __setitem__()
        # to pass certain options directly to the canvas.
        if cnf:
            for key in cnf:
                self[key] = cnf[key]

        for key in kw:
            self[key] = kw[key]

    # Also override this alias for configure()
    config = configure

    def display_widget(self, widget_class, fit_width=False, **kw):
        """Create and display a new widget.

        If fit_width == True, the interior widget will be stretched as
        needed to fit the width of the frame.

        Keyword arguments are passed to the widget_class constructor.

        Returns the new widget.
        """

        # Blank the canvas
        self.erase()

        # Set width fitting
        self._fit_width = fit_width

        # Set the new interior widget
        self._interior = widget_class(self._canvas, **kw)

        # Add the interior widget to the canvas, and save its widget ID
        # for use in _resize_interior()
        self._interior_id = self._canvas.create_window(0, 0,
                                                       anchor="nw",
                                                       window=self._interior)

        # Call _update_scroll_region() when the interior widget is resized
        self._interior.bind("<Configure>", self._update_scroll_region)

        # Fit the interior widget to the canvas if requested
        # We don't need to check fit_width here since _resize_interior()
        # already does.
        self._resize_interior()

        # Scroll to the top-left corner of the canvas
        self.scroll_to_top()

        return self._interior

    def erase(self):
        """Erase the displayed widget."""

        # Clear the canvas
        self._canvas.delete("all")

        # Delete the interior widget
        del self._interior
        del self._interior_id

        # Save these names
        self._interior = None
        self._interior_id = None

        # Reset width fitting
        self._fit_width = False

    def scroll_to_top(self):
        """Scroll to the top-left corner of the canvas."""

        self._canvas.xview_moveto(0)
        self._canvas.yview_moveto(0)

    # ------------------------------------------------------------------------

    def _resize_interior(self, event=None):
        """Resize the interior widget to fit the canvas."""

        if self._fit_width and self._interior_id:
            # The current width of the canvas
            canvas_width = self._canvas.winfo_width()

            # The interior widget's requested width
            requested_width = self._interior.winfo_reqwidth()

            if requested_width != canvas_width:
                # Resize the interior widget
                new_width = max(canvas_width, requested_width)
                self._canvas.itemconfigure(self._interior_id, width=new_width)

    def _scroll_canvas(self, event):
        """Scroll the canvas."""

        c = self._canvas

        if sys.platform.startswith("darwin"):
            # macOS
            c.yview_scroll(-1 * event.delta, "units")

        elif event.num == 4:
            # Unix - scroll up
            c.yview_scroll(-1, "units")

        elif event.num == 5:
            # Unix - scroll down
            c.yview_scroll(1, "units")

        else:
            # Windows
            c.yview_scroll(-1 * (event.delta // 120), "units")

    def _update_scroll_region(self, event):
        """Update the scroll region when the interior widget is resized."""

        # The interior widget's requested width and height
        req_width = self._interior.winfo_reqwidth()
        req_height = self._interior.winfo_reqheight()

        # Set the scroll region to fit the interior widget
        self._canvas.configure(scrollregion=(0, 0, req_width, req_height))

    # ------------------------------------------------------------------------

    # Keys for configure() to forward to the canvas widget
    _CANVAS_KEYS = "width", "height", "takefocus", "bg"

    # Scrollbar-related configuration
    _DEFAULT_SCROLLBARS = "both"
    _VALID_SCROLLBARS = "vertical", "horizontal", "both", "neither"



# ------------------------------------------------------------------------------
#  File: gamma_correction.py
#  Author: Jan Kukacka
#  Date: 2/2022
# ------------------------------------------------------------------------------
#  Implementation of Gamma correction
# ------------------------------------------------------------------------------

import numpy as np

try:
    from . import filter
except ImportError:
    from filters import filter

class GammaCorrection(filter.Filter):
    '''
    Performs gamma correction (power-law)
    '''

    name = 'gamma_correction'

    def __init__(self, gamma=1):
        '''
        # Arguments:
            - gamma: Power to apply
        '''
        super().__init__()
        self.gamma = gamma

    def __call__(self, img):
        '''
        # Arguments:
            - img: array of data to be normalized

        # Returns:
            - normalized array.
        '''
        result = super().__call__(img)
        if result is not None:
            return result
        else:
            self.cache = self.call(img, self.gamma)
            return self.cache


    @staticmethod
    def call(img, gamma, **kwargs):
        img_min, img_max = img.min(), img.max()

        norm_img = np.power(img, gamma)

        ## Power stays within [0;1] range
        if img_min == 0 and img_max == 1:
            return norm_img
        else:
            norm_min, norm_max = norm_img.min(), norm_img.max()
            scale = (img_max-img_min) / (norm_max-norm_min)
            return img_min + (norm_img-norm_min) * scale


    def serialize(self):
        base_dict = super().serialize()
        base_dict['params']['gamma'] = self.gamma
        return base_dict

    @staticmethod
    def deserialize(serialization):
        gamma = serialization['gamma']
        obj = GammaCorrection(gamma)
        obj._deserialize_parent(serialization)
        return obj



# ------------------------------------------------------------------------------
#  File: tk_lazy_var.py
#  Author: Jan Kukacka
#  Date: 2/2022
# ------------------------------------------------------------------------------
#  Lazy variables that only trigger 'write' trace events when their value
#  really changes
# ------------------------------------------------------------------------------

import tkinter as tk

class StringVar(tk.StringVar):
    def set(self, val):
        if val != self.get():
            super().set(val)

class BooleanVar(tk.BooleanVar):
    def set(self, val):
        if val != self.get():
            super().set(val)

class DoubleVar(tk.DoubleVar):
    def set(self, val):
        if val != self.get():
            super().set(val)

class IntVar(tk.IntVar):
    def set(self, val):
        if val != self.get():
            super().set(val)



# ------------------------------------------------------------------------------
#  File: observabledict.py
#  Author: Remi Emonet
#  Date: 8/2019
#  Source: https://github.com/twitwi/vuejs-python/tree/master/vuejspython/observablecollections
#  License: MIT
# ------------------------------------------------------------------------------

from .event import Event
from .observable import Observable
from collections import namedtuple

class ObservableDict(dict, Observable):
	_itemTuple = namedtuple('item', 'key value')
	_itemChangedTuple = namedtuple('item', 'key value oldValue')

	def __init__(self, *args, **kwargs):
		Observable.__init__(self)
		dict.__init__(self, *args, **kwargs)

	def popitem(self):
		keyValue = dict.popitem(self)
		self.raiseEvent('itemsRemoved', items=[self._createItemTuple(keyValue[0], keyValue[1])])
		return keyValue

	def pop(self, key, default=None):
		value = dict.pop(self, key, default)
		self.raiseEvent('itemsRemoved', items=[self._createItemTuple(key, value)])
		return value

	def __setitem__(self, key, value):
		oldValue = dict.get(self, key)

		new_is_different = True
		try:
			if oldValue == value:
				new_is_different = False
		except Exception as e:
			print(e, value)
			pass
		if new_is_different:
			dict.__setitem__(self, key, value)
			item = self._createItemTuple(key, value, oldValue)

			if (oldValue != None):
				actionName = 'itemsUpdated'
			else:
				actionName = 'itemsAdded'

			self.raiseEvent(actionName, items=[item])

	def __delitem__(self, key):
		value = dict.__getitem__(self, key)
		dict.__delitem__(self, key)
		self.raiseEvent('itemsRemoved', items=[ self._createItemTuple(key, value) ])

	def update(self, dict2):
		itemsAdded = []
		itemsChanged = []

		for keyValue in dict2.items():
			key = keyValue[0]
			value = keyValue[1]

			currValue = dict.get(self, key)

			if currValue == None:
				dict.__setitem__(self, key, value)
				itemsAdded.append(self._createItemTuple(key, value))
			elif currValue != value:
				dict.__setitem__(self, key, value)
				itemsChanged.append(self._createItemTuple(key, value, oldValue))

		if len(itemsAdded) > 0:
			self.raiseEvent('itemsAdded', items=itemsAdded)

		if len(itemsChanged) > 0:
			self.raiseEvent('itemsUpdated', items=itemsChanged)

	def clear(self):
		itemsRemoved = []
		items = self.items()

		for item in items:
			itemsRemoved.append(self._createItemTuple(item[0], item[1]))
			dict.__delitem__(self, item[0])

		self.raiseEvent('itemsRemoved', items=itemsRemoved)

	def _createItemTuple(self, key, value, oldValue=None):

		if oldValue == None:
			item = self._itemTuple(key=key, value=value)
		else:
			item = self._itemChangedTuple(key=key, value=value, oldValue=oldValue)
		return item



# ------------------------------------------------------------------------------
#  File: renderer.py
#  Author: Jan Kukacka
#  Date: 11/2021
# ------------------------------------------------------------------------------
#  Image rendering thread code
# ------------------------------------------------------------------------------

import traceback
import numpy as np
import happy as hp
from PIL import Image
from time import time
from queue import Empty
from functools import reduce
from multiprocessing import Process, Queue
from matplotlib.colors import PowerNorm

try:
    from .filters.pipeline import Pipeline
except ImportError:
    from filters.pipeline import Pipeline


def render(rendering_queue, rendered_queue, use_gpu, debug, drop_tasks=True):
    '''
    Code for the rendering process
    '''
    image_local = None
    image_local_changed = False
    pipelines = {}
    cache = {}
    while True:
        ## Flush old tasks, work only on the last one
        ## NOTE: This is only reliable with a single consumer thread
        task = rendering_queue.get()
        if task is not None and 'image' in task:
            image_local = task['image']
            image_local_changed = True

        if drop_tasks:
            try:
                while True:
                    task = rendering_queue.get(False)
                    # task = rendering_queue.get(True, .05)
                    if task is not None and 'image' in task:
                        image_local = task['image']
                        image_local_changed = True
                    # print('Dropping a task')
            except Empty:
                pass

        ## Termination signal
        if task is None:
            # print('Exiting rendering thread')
            break

        time_render = 0
        time_validation = 0
        time_coloring = 0
        t0 = time()
        try:
            processed_images = []
            response_images = []
            if image_local_changed:
                pipelines = {}
                colors = {}
                image_local_changed = False


            for channel_index, channel_property in enumerate(task['channel_properties']):
                ## Ignore hidden channels
                if not channel_property['visible']:
                    bkg = np.zeros((128,256,4), dtype=np.uint8)
                    bkg[::32] = bkg[-1] = bkg[:,::32] = bkg[:,-1] = 0x66
                    response_images.append(bkg)
                    continue

                image = image_local[...,channel_index]
                t1 = time()
                if (channel_index not in pipelines
                    or pipelines[channel_index].update(channel_property['pipeline'])
                    or channel_property['color'] != colors[channel_index]):

                    if channel_index not in pipelines:
                        pipelines[channel_index] = Pipeline.deserialize(channel_property['pipeline'])
                    t2 = time()
                    mn,mx = image.min(), image.max()
                    image = (image-mn)/(mx-mn)
                    output_image = pipelines[channel_index](image)

                    colors[channel_index] = channel_property['color']
                    cmap = hp.plots.cmap((0,'#444444'),(1/256, 'k'), (1,channel_property['color']))
                    response_image = render_response(image, output_image, cmap)
                    t3 = time()
                    output_image = hp.plots.cmap('k', channel_property['color'])(output_image)
                    t4 = time()
                    cache[channel_index] = output_image, response_image
                else:
                    t2 = time()
                    output_image, response_image = cache[channel_index]
                    t4 = t3 = time()

                time_render += t3-t2
                time_validation += t2-t1
                time_coloring += t4-t3
                processed_images.append(output_image)
                response_images.append(response_image)

            ## Render
            t5 = time()
            ## NOTE: reduce is faster here than stacking the list of arrays and
            ##       calling np.sum on them
            render = reduce(np.add, processed_images)
            render = np.minimum(render, 1) * 255
            render = Image.fromarray(render.astype(np.uint8))
            t6 = time()
            # if debug:
            #     print(f'Pipeline validation: {time_validation:.3f} Rendering: {time_render:.3f} Coloring: {time_coloring:.3f} Sum: {t6-t5:.3f} Total: {t6-t0:.3f}')
            rendered_queue.put((render, response_images))
        except Exception as e:
            if debug:
                track = traceback.format_exc()
                print('Error in Rendering Thread:')
                print(track)

    ## Signal finish of the rendered queue before quitting - it needs to be emptied
    rendered_queue.put(None)


def render_response(input_image, output_image, cmap):
    from skimage.transform import resize
    response = np.empty((128,256,4))
    response[:] = cmap(np.linspace(0,1,response.shape[0]))[:,None]
    hist = np.histogram2d(output_image.ravel(), input_image.ravel(), bins=(np.linspace(0,1,64),np.linspace(0,1,128)))[0]
    hist = np.log(hist)
    hist = np.nan_to_num(0.5 + 0.5*(hist / hist.max()), neginf=0)
    response[...,-1] = resize(hist, response.shape[:2], preserve_range=True)
    response = (255*response).astype(np.uint8)[::-1]
    bkg = np.zeros_like(response)
    bkg[::32] = bkg[-1] = bkg[:,::32] = bkg[:,-1] = 0x66
    mask = response[:,:,-1] == 0
    response[mask] = bkg[mask]
    return response



# ------------------------------------------------------------------------------
#  File: event.py
#  Author: Remi Emonet
#  Date: 8/2019
#  Source: https://github.com/twitwi/vuejs-python/tree/master/vuejspython/observablecollections
#  License: MIT
# ------------------------------------------------------------------------------

class Event:
	def __init__(self, action, source):
		self._action = action
		self._source = source

	@property
	def action(self):
		return self._action

	@property
	def source(self):
		return self._source



# ------------------------------------------------------------------------------
#  File: filter_factory.py
#  Author: Jan Kukacka
#  Date: 2/2022
# ------------------------------------------------------------------------------
#  Filter factory
# ------------------------------------------------------------------------------

try:
    from . import local_norm
    from . import sigmoid_norm
    from . import unsharp_mask
    from . import gamma_correction
    from . import frangi
    from . import minmax_norm
    from . import gaussian_blur
    from . import anisotropic_denoising
except ImportError:
    from filters import local_norm
    from filters import sigmoid_norm
    from filters import unsharp_mask
    from filters import gamma_correction
    from filters import frangi
    from filters import minmax_norm
    from filters import gaussian_blur
    from filters import anisotropic_denoising

__available_filters = (local_norm.LocalNorm,
                       sigmoid_norm.SigmoidNorm,
                       unsharp_mask.UnsharpMask,
                       gamma_correction.GammaCorrection,
                       frangi.Frangi,
                       minmax_norm.MinMaxNorm,
                       gaussian_blur.GaussianBlur,
                       anisotropic_denoising.AnisotropicDenoising)

def get_filter_by_name(name):
    '''
    Return type matching given name
    '''
    for T_filter in __available_filters:
        if name == T_filter.name:
            return T_filter

def get_available_filters():
    return __available_filters



# ------------------------------------------------------------------------------
#  File: window_about.py
#  Author: Jan Kukacka
#  Date: 11/2021
# ------------------------------------------------------------------------------
#  Implementation of the About window
# ------------------------------------------------------------------------------

import tkinter as tk
import tkinter.ttk as ttk
import os
from pathlib import Path
from PIL import Image, ImageOps, ImageTk
import numpy as np

try:
    from importlib import metadata
except ImportError: # for Python<3.8
    import importlib_metadata as metadata
__version__ = metadata.version('image-viewer-mk2')


def prepare_icon(filename, skin):
    path = Path(os.path.dirname(os.path.abspath(__file__)))
    icon = Image.open(str(path.parent/'resources'/filename))
    icon2 = icon.convert('RGB')
    icon2 = ImageOps.invert(icon2)
    icon2 = Image.fromarray(np.concatenate([np.array(icon2), np.array(icon)[...,-1,None]], axis=-1), mode='RGBA')
    bkg = Image.new(mode='RGBA', size=icon.size, color=skin.bg_color)
    icon = Image.alpha_composite(bkg, icon2).convert('RGB')
    icon = ImageTk.PhotoImage(icon)
    return icon


class WindowAbout(tk.Toplevel):

    def __init__(self, master, skin):
        tk.Toplevel.__init__(self, master)
        self.master = master
        self.skin = skin

        self.title('About')
        self.mainframe = tk.Frame(self, background=self.skin.bg_color)
        self.mainframe.pack(side=tk.BOTTOM, fill=tk.BOTH, expand=True)
        ## Hide window on closing
        self.protocol("WM_DELETE_WINDOW", self.on_closing)

        frame = tk.Frame(self.mainframe, bg=self.skin.bg_color)
        frame.grid(column=0, row=0, sticky='nwes')

        icon = prepare_icon('logo128.png', self.skin)
        tk.Label(frame, image=icon, bg=self.skin.bg_color).pack(side=tk.TOP, pady=100, padx=30)
        self.icon = icon

        frame = tk.Frame(self.mainframe, bg=self.skin.bg_color)
        frame.grid(column=1, row=0, sticky='nwes')

        frame = tk.Frame(frame, bg='#ff0000')
        frame.pack(side=tk.TOP, expand=True, fill=tk.X, padx=10)
        tk.Label(frame, text='Image Viewer MK2', fg=self.skin.fg_color, bg=self.skin.bg_color, font=('Segoe UI', 14, 'bold')).pack(side=tk.TOP, expand=False, fill=tk.X)
        tk.Label(frame, text=f'v{__version__} (2022-02-07)', fg=self.skin.fg_color, bg=self.skin.bg_color).pack(side=tk.TOP, expand=False, fill=tk.X)
        tk.Label(frame, text='Author: Jan Kukačka, 2021', fg=self.skin.fg_color, bg=self.skin.bg_color).pack(side=tk.TOP, expand=False, fill=tk.X)
        tk.Label(frame, text='Provided under MIT license.', fg=self.skin.fg_color, bg=self.skin.bg_color).pack(side=tk.TOP, expand=False, fill=tk.X)
        tk.Label(frame, text='Icon credits: Icon home, Gregor Cresnar,\nFreepik, Google, Uptal Barman\nArkinasi, Royyan Wijaya and Pancracysdh.', fg=self.skin.fg_color, bg=self.skin.bg_color).pack(side=tk.TOP, expand=False, fill=tk.X)

    def on_closing(self):
        self.master.window_about = None
        self.withdraw()



# ------------------------------------------------------------------------------
#  File: frangi.py
#  Author: Jan Kukacka
#  Date: 2/2022
# ------------------------------------------------------------------------------
#  Implementation of Vesselness filter
# ------------------------------------------------------------------------------

import numpy as np
from skimage.filters import frangi, meijering

try:
    from . import filter
except ImportError:
    from filters import filter

class Frangi(filter.Filter):
    '''
    Applies Frangi filter
    '''

    name = 'frangi'

    def __init__(self, scale_min=1, scale_max=10, scale_step=2, alpha=0.5, beta=.5, gamma=15):
        '''
        # Arguments:
            - see parameters of skimage.filters.frangi.
        '''
        super().__init__()
        self.scale_min = scale_min
        self.scale_max = scale_max
        self.scale_step = scale_step
        self.alpha = alpha
        self.beta = beta
        self.gamma = gamma

    def __call__(self, img):
        '''
        # Arguments:
            - img: array of data to be normalized

        # Returns:
            - normalized array.
        '''
        result = super().__call__(img)
        if result is not None:
            return result
        else:
            self.cache = self.call(img, self.scale_min, self.scale_max, self.scale_step, self.alpha, self.beta, self.gamma)
            return self.cache


    @staticmethod
    def call(img, scale_min, scale_max, scale_step, alpha, beta, gamma, **kwargs):
        img_min, img_max = img.min(), img.max()

        sigmas = np.arange(min(scale_min, scale_max), max(scale_min, scale_max), scale_step)
        # norm_img = frangi(img, sigmas=sigmas, alpha=alpha, beta=beta, gamma=gamma, black_ridges=False)
        norm_img = meijering(img, sigmas=sigmas, alpha=alpha, black_ridges=False)

        norm_min, norm_max = norm_img.min(), norm_img.max()
        scale = (img_max-img_min) / (norm_max-norm_min)
        return img_min + (norm_img-norm_min) * scale


    def serialize(self):
        base_dict = super().serialize()
        base_dict['params']['scale_min'] = self.scale_min
        base_dict['params']['scale_max'] = self.scale_max
        base_dict['params']['scale_step'] = self.scale_step
        base_dict['params']['alpha'] = self.alpha
        base_dict['params']['beta'] = self.beta
        base_dict['params']['gamma'] = self.gamma
        return base_dict

    @staticmethod
    def deserialize(serialization):
        scale_min = serialization['scale_min']
        scale_max = serialization['scale_max']
        scale_step = serialization['scale_step']
        alpha = serialization['alpha']
        beta = serialization['beta']
        gamma = serialization['gamma']
        obj = Frangi(scale_min, scale_max, scale_step, alpha, beta, gamma)
        obj._deserialize_parent(serialization)
        return obj



# ------------------------------------------------------------------------------
#  File: view.py
#  Author: Jan Kukacka
#  Date: 3/2021
# ------------------------------------------------------------------------------
#  Main window view
# ------------------------------------------------------------------------------

import os
# import matplotlib
# import matplotlib.colors
# import matplotlib.lines as mlines
# import matplotlib.pyplot as plt
# from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk
import numpy as np
from pathlib import Path
from PIL import ImageTk, Image
import tkinter as tk
import tkinter.ttk as ttk
from tkinter.colorchooser import askcolor as askcolor_
from tkinter.filedialog import askopenfilename as askopenfilename_
from tkinter.filedialog import asksaveasfilename as asksaveasfilename_


import happy as hp
import happy.plots as hpp

try:
    from . import skin as skin_
    from .tk_widgets.limiter import Limiter
    from .tk_widgets.loader_animation import LoaderAnimation
    from .tk_widgets.channels_list import ChannelsList
    from .tk_widgets.panel_pipelines import PanelPipelines
    from .tk_widgets.window_about import WindowAbout
    from .utils.tk_call_wrapper import CallWrapper
    from .tk_widgets import filter_config
    from .tk_widgets import sortable_accordion
    from .filters.filter_factory import get_available_filters
    from .utils.tk_lazy_var import StringVar, BooleanVar
except ImportError:
    import skin as skin_
    from tk_widgets.limiter import Limiter
    from tk_widgets.loader_animation import LoaderAnimation
    from tk_widgets.channels_list import ChannelsList
    from tk_widgets.window_about import WindowAbout
    from utils.tk_call_wrapper import CallWrapper
    from tk_widgets import filter_config
    from tk_widgets import sortable_accordion
    from filters.filter_factory import get_available_filters
    from utils.tk_lazy_var import StringVar, BooleanVar

def set_state(element, state):
    try:
        element.configure(state=state)
    except:
        try:
            ## ttk widgets use different method
            if state == 'disable':
                element.state(['disabled'])
            if state == 'normal':
                element.state(['!disabled'])
        except:
            pass

    for child in element.winfo_children():
        set_state(child, state)


class View(tk.Tk):
    '''
    Main window view class
    '''

    def __init__(self, skin=None, debug=None, *args, **kwargs):
        tk.Tk.__init__(self, *args, **kwargs)

        self.zoom = 1
        self.offset = (0,0)

        if skin is None:
            skin = skin_.Skin()
        self.skin = skin

        if debug is None:
            debug = False
        if not debug:
            tk.CallWrapper = CallWrapper

        self.window_about = None

        self.setup_mainframe()
        self.setup_menu()
        self.setup_image_axis()
        self.setup_response_panel()
        self.setup_channels_panel()
        self.setup_channels2_panel()
        self.setup_model_io()

        self.loader = LoaderAnimation((self.figure_canvas.winfo_width()/2, 1), self.skin.bg_color, self.figure_canvas, self)

        ## -- bind mouse on canvas
        self.figure_canvas.bind('<Button-1>', self.mouse1_drag)
        self.figure_canvas.bind('<B1-Motion>', self.mouse1_drag)
        self.figure_canvas.bind('<Button-3>', self.mouse2_drag)
        self.figure_canvas.bind('<B3-Motion>', self.mouse2_drag)

        ## -- bind keyboard
        self.bind('+', lambda _,self=self:self.set_zoom(self.zoom+.5))
        self.bind('-', lambda _,self=self:self.set_zoom(self.zoom-.25))
        self.bind('r', self.reset_view)

        self.title('Image Viewer MK II')
        path = Path(os.path.dirname(os.path.abspath(__file__)))
        self.iconbitmap(True, str(path/'resources'/'favicon.ico'))


    def setup_mainframe(self):
        self.mainframe = tk.Frame(self, background=self.skin.bg_color)
        self.mainframe.pack(side=tk.BOTTOM, fill=tk.BOTH, expand=True)
        self.mainframe.columnconfigure(0, weight=1)
        self.mainframe.rowconfigure(0, weight=1)

        self.grid_frames = []
        for i in range(3):
            frame = tk.Frame(self.mainframe, background=self.skin.bg_color)
            frame.grid(column=i, row=0, sticky='nwes')
            self.grid_frames.append(frame)

    def setup_menu(self):
        self.menu = {'obj': tk.Menu(self.mainframe)}
        self.menu['file'] = {'obj': tk.Menu(self.menu['obj'])}
        self.menu['file']['load_image'] = 'Load image... (Ctrl+O)'
        self.menu['file']['obj'].add_command(label=self.menu['file']['load_image'])
        self.menu['file']['save_render'] = 'Save image... (Ctrl+S)'
        self.menu['file']['obj'].add_command(label=self.menu['file']['save_render'])
        self.menu['file']['load_config'] = 'Load configuration'
        self.menu['file']['obj'].add_command(label=self.menu['file']['load_config'])
        self.menu['file']['save_config'] = 'Save configuration'
        self.menu['file']['obj'].add_command(label=self.menu['file']['save_config'])
        self.menu['obj'].add_cascade(label="File", menu=self.menu['file']['obj'])

        self.menu['image'] = {'obj': tk.Menu(self.menu['obj'])}
        self.menu['image']['transpose'] = 'Transpose (Ctrl+T)'
        self.menu['image']['obj'].add_command(label=self.menu['image']['transpose'])
        self.menu['image']['autocolor'] = 'Default palette'
        self.menu['image']['obj'].add_command(label=self.menu['image']['autocolor'])
        self.menu['obj'].add_cascade(label="Image", menu=self.menu['image']['obj'])

        self.menu['view'] = {'obj': tk.Menu(self.menu['obj'])}
        self.menu['view']['reset'] = 'Reset (R)'
        self.menu['view']['obj'].add_command(label=self.menu['view']['reset'], command=self.reset_view)
        self.menu['view']['zoomin'] = 'Zoom in (+)'
        self.menu['view']['obj'].add_command(label=self.menu['view']['zoomin'], command=lambda self=self:self.set_zoom(self.zoom+.5))
        self.menu['view']['zoomout'] = 'Zoom out (-)'
        self.menu['view']['obj'].add_command(label=self.menu['view']['zoomout'], command=lambda self=self:self.set_zoom(self.zoom-.25))
        self.menu['obj'].add_cascade(label="View", menu=self.menu['view']['obj'])

        self.menu['about'] = 'About'
        self.menu['obj'].add_command(label=self.menu['about'], command=self.open_about)

        tk.Tk.config(self, menu=self.menu['obj'])


    def setup_image_axis(self):
        self.figure_canvas = tk.Canvas(self.grid_frames[0], width=500, height=500, bg=self.skin.bg_color, bd=0, highlightthickness=0)
        self.figure_canvas.pack(side=tk.BOTTOM, fill=tk.BOTH, expand=True)

    def setup_channels_panel(self):
        self.property_frames = {}
        # self.channel_figures = [None,None]

        self.var_channel = {
            'color': StringVar(value='#000000'),
            'visible': BooleanVar(value=True)
        }

        channel_frame = ttk.LabelFrame(self.grid_frames[2], text='Channel settings', padding=5)
        channel_frame.pack(side=tk.TOP, expand=False, fill=tk.BOTH, padx=5, pady=5)

        self.var_selected_channel = StringVar(value='Channel 0')
        self.channel_combo = ttk.Combobox(channel_frame, state='readonly',
                             textvariable=self.var_selected_channel)
        self.channel_combo.pack(side=tk.TOP, expand=False, fill=tk.X)


        color_frame = tk.Frame(channel_frame, bg=self.skin.bg_color)
        color_frame.pack(side=tk.TOP, expand=True, fill=tk.BOTH, padx=0, pady=5)

        grid = tk.Frame(color_frame, bg=self.skin.bg_color)
        grid.pack(side=tk.TOP, expand=True, fill=tk.X)
        self.color_preview = tk.Frame(grid, width=16, height=16, bg=self.var_channel['color'].get(), highlightthickness=1, highlightbackground='#000000')
        self.color_preview.pack(side=tk.LEFT)

        ttk.Entry(grid, textvariable=self.var_channel['color'], width=10).pack(side=tk.LEFT, padx=5)
        path = Path(os.path.dirname(os.path.abspath(__file__)))
        colorpicker = tk.PhotoImage(file=str(path/'resources'/'colorpicker16.png'))
        self.color_picker = ttk.Button(grid, image=colorpicker)
        self.color_picker.image = colorpicker
        self.color_picker.pack(side=tk.LEFT)

        icon = tk.PhotoImage(file=str(path/'resources'/'paste16.png'))
        self.btn_paste = ttk.Button(grid, image=icon)
        self.btn_paste.image = icon
        self.btn_paste.pack(side=tk.RIGHT)

        icon = tk.PhotoImage(file=str(path/'resources'/'copy16.png'))
        self.btn_copy = ttk.Button(grid, image=icon)
        self.btn_copy.image = icon
        self.btn_copy.pack(side=tk.RIGHT, padx=3)

        self.pipelines_panel = PanelPipelines(channel_frame, self.skin, self.var_selected_channel)

        self.btn_add_filter = ttk.Button(channel_frame, text='Add filter', command=self.show_filter_menu)
        self.btn_add_filter.pack(side=tk.BOTTOM)
        self.menu_add_filter = {'obj': tk.Menu(self)}
        for T_filter in get_available_filters():
            T_widget = filter_config.get_filter_widget(T_filter.name)
            if T_widget is not None:
                self.menu_add_filter['obj'].add_command(label=T_widget.title)
                self.menu_add_filter[T_filter.name] = T_widget.title


    def setup_channels2_panel(self):
        self.channels_panel = ChannelsList(self.grid_frames[1],self.skin, self.var_selected_channel)


    def setup_response_panel(self):
        frame = ttk.LabelFrame(self.grid_frames[2], text='Channel response', padding=5)
        frame.pack(side=tk.TOP, expand=False, fill=tk.BOTH, padx=5, pady=5)
        self.response_canvas = tk.Canvas(frame, width=256, height=128, bg=self.skin.bg_color, bd=0, highlightthickness=0)
        self.response_canvas.pack(side=tk.BOTTOM, fill=tk.BOTH, expand=True)


    def setup_model_io(self):
        btns_frame = tk.Frame(self.grid_frames[2], bg=self.skin.bg_color)
        btns_frame.pack(side=tk.TOP, pady=5, padx=5, expand=False, fill=tk.X)
        self.btn_save = ttk.Button(btns_frame, text='Save config')
        self.btn_save.pack(side=tk.LEFT, anchor=tk.NE)
        self.btn_load = ttk.Button(btns_frame, text='Load config')
        self.btn_load.pack(side=tk.LEFT, padx=5, anchor=tk.NE)
        self.btn_save_render = ttk.Button(btns_frame, text='Save render')
        self.btn_save_render.pack(side=tk.LEFT, anchor=tk.NE)

    @staticmethod
    def askcolor(initialcolor):
        return askcolor_(initialcolor)[1]

    @staticmethod
    def askopenfilename(*args, **kwargs):
        return askopenfilename_(*args, **kwargs)

    @staticmethod
    def asksaveasfilename(*args,**kwargs):
        return asksaveasfilename_(*args, **kwargs)


    def show_image(self, image=None):
        if image is not None:
            self.image = image

        if self.zoom != 1:
            image = self.image.resize((int(self.image.size[0]*self.zoom),
                                       int(self.image.size[1]*self.zoom)))
        else:
            image = self.image
        self.render_ref = ImageTk.PhotoImage(image)
        # self.figure_canvas.update()
        canvas_width = self.figure_canvas.winfo_width()
        canvas_height = self.figure_canvas.winfo_height()
        # print(canvas_width, canvas_height)
        self.figure_canvas.create_image((canvas_width/2+self.offset[0],canvas_height/2+self.offset[1]), image=self.render_ref)

    def show_response(self, response_image):
        if not response_image is None:
            self.response_ref = ImageTk.PhotoImage(image=Image.fromarray(response_image))
            self.response_canvas.create_image((0,0), image=self.response_ref,anchor=tk.NW)

    def show_filter_menu(self):
        try:
            x = self.btn_add_filter.winfo_rootx()
            y = self.btn_add_filter.winfo_rooty()
            self.menu_add_filter['obj'].tk_popup(x, y, 0)
        finally:
            self.menu_add_filter['obj'].grab_release()

    def update_channels(self, channel_props):
        self.channel_combo.configure(values=[cp['name'] for cp in channel_props])
        self.channels_panel.recreate_items(channel_props)
        self.pipelines_panel.recreate_items(channel_props)

    def get_active_channel(self):
        channel_names = [vc['name'].get() for vc in self.channels_panel.var_channels]
        if self.var_selected_channel.get() in channel_names:
            return channel_names.index(self.var_selected_channel.get())
        else:
            return -1

    def mouse1_drag(self, event):
        '''
        Changes image offset
        '''
        if event.type == tk.EventType.Button:
            self.offset_root = (event.x, event.y)
            self.orig_offset = self.offset
        else:
            self.offset = (event.x - self.offset_root[0] + self.orig_offset[0],
                           event.y - self.offset_root[1] + self.orig_offset[1])
            self.show_image()

    def mouse2_drag(self, event):
        '''
        Changes image zoom
        '''
        if event.type == tk.EventType.Button:
            self.orig_zoom = self.zoom
            self.zoom_root = (event.x, event.y)
        else:
            distance = (self.zoom_root[1] - event.y) / 200
            activation = 1 + max(0, distance) + min(0, distance/2)
            self.set_zoom(self.orig_zoom * activation)

    def set_zoom(self, zoom):
        self.zoom = np.maximum(np.minimum(8, zoom),1/2)
        self.show_image()

    def reset_view(self, *args):
        self.zoom = 1
        self.offset = (0,0)
        self.show_image()

    def open_about(self):
        if self.window_about is None:
            self.window_about = WindowAbout(self, self.skin)
            self.window_about.deiconify()



# ------------------------------------------------------------------------------
#  File: loader_animation.py
#  Author: Jan Kukacka
#  Date: 3/2021
# ------------------------------------------------------------------------------
#  Class that implements loader animation in the view
# ------------------------------------------------------------------------------

import os
from PIL import Image, ImageTk, ImageSequence, ImageChops, ImageEnhance
from pathlib import Path

class LoaderAnimation(object):

    def __init__(self, position, bg_color, tk_canvas, tk):
        '''
        # Arguments:
        - `bg_color`: background color of the loader animation.
        '''
        self.position = position
        self.bg_color = bg_color
        self.tk_canvas = tk_canvas
        self.tk = tk

        path = Path(os.path.dirname(os.path.abspath(__file__))).parent / 'resources' / 'loader.gif'
        img = Image.open(path)

        self.prepare_frames(img)

        self._is_playing = False
        self._active_frame = 0
        self._dim_level = 0

    def prepare_frames(self, img):
        self.frames = [im.copy().crop((130,120,270,183)) for im in ImageSequence.Iterator(img)]
        for i in range(len(self.frames)):
            bkg = Image.new("RGB", self.frames[i].size, self.bg_color)
            frame = ImageChops.add(bkg, ImageChops.invert(self.frames[i].convert('RGB')))
            frame = frame.convert("HSV")

            # adjust = Image.new("HSV", frame.size, (140,0,0))
            # frame = ImageChops.add_modulo(frame, adjust)
            # adjust = Image.new("HSV", frame.size, (0,0,0))
            # frame = ImageChops.add(frame, adjust)
            frame = frame.convert("RGBA")

            gray = ImageChops.invert(self.frames[i].convert('L'))
            E1 = ImageEnhance.Contrast(gray)

            frame.putalpha(E1.enhance(2))#ImageChops.multiply(gray, gray))
            # bkg = Image.new("RGB", frame.size, self.bg_color)
            # self.frames[i] = ImageChops.screen(bkg, ImageChops.invert(frame))
            self.frames[i] = frame

    def show(self):
        if not self._is_playing:
            self._is_playing = True
        if self._dim_level <= 0:
            self.tk.after(10, self._update)

    def hide(self):
        self._is_playing = False

    def _update(self):
        frame = self.frames[self._active_frame]
        if self._dim_level < 1:
            frame = frame.copy()
            dim = frame.getchannel('A')
            E = ImageEnhance.Brightness(dim)
            frame.putalpha(E.enhance(self._dim_level))

        self._render_ref = ImageTk.PhotoImage(frame)
        self.tk_canvas.create_image((self.tk_canvas.winfo_width()/2, self.tk_canvas.winfo_height()/2), image=self._render_ref)
        self._active_frame = (self._active_frame + 1) % len(self.frames)

        if self._is_playing or self._dim_level > 0:
            self.tk.after(20, self._update)

        if self._is_playing and self._dim_level < 1:
            self._dim_level += 0.05
        if not self._is_playing and self._dim_level > 0:
            self._dim_level -= 0.01



# ------------------------------------------------------------------------------
#  File: utils.py
#  Author: Jan Kukacka
#  Date: 2/2022
# ------------------------------------------------------------------------------
#  Helping functions for working with observable collections
# ------------------------------------------------------------------------------

try:
    from .observablelist import ObservableList
    from .observabledict import ObservableDict
except ImportError:
    from ObservableCollections.observablelist import ObservableList
    from ObservableCollections.observabledict import ObservableDict


def make_observable(obj):
    if isinstance(obj, list):
        return ObservableList([make_observable(item) for item in obj])
    elif isinstance(obj, dict):
        return ObservableDict({key:make_observable(val) for key,val in obj.items()})
    elif isinstance(obj, set):
        raise NotImplementedError('make_observable does not support sets.')
    else:
        return obj

def make_plain(obj):
    if isinstance(obj, (list,ObservableList)):
        return [make_plain(item) for item in obj]
    if isinstance(obj, tuple):
        return tuple(make_plain(item) for item in obj)
    elif isinstance(obj, (dict,ObservableDict)):
        return {key:make_plain(val) for key,val in obj.items()}
    # elif isinstance(obj, ObservableSet):
    #     raise NotImplementedError('make_plain does not support sets.')
    else:
        return obj



# ------------------------------------------------------------------------------
#  File: app.py
#  Author: Jan Kukacka
#  Date: 3/2021
# ------------------------------------------------------------------------------
#  App interface - processes inputs, starts GUI, returns output
# ------------------------------------------------------------------------------

import numpy as np
import happy as hp
try:
    from . import view as view_
    from . import model as model_
    from . import presenter as presenter_
except ImportError:
    import view as view_
    import model as model_
    import presenter as presenter_


def start(file=None, image=None, **kwargs):
    '''
    Starts the app and opens optionally given file or uses given image.

    # Arguments:
    - file: Optional. Filename to open
    - image: Optional. Alternatively, an opened image can be passed as an array
        of shape (height, width, n_channels)

    # Additional optional kwargs:
    - gpu: bool. If True, PyTorch+GPU based rendering will be used (if
        installed). If False (default), defaults to NumPy+CPU rendering
    - config_filename: Filename of the config to apply.
    - config: Dictionary with config to apply.
    - return_config: bool. If True, returns also the config dict. False by default.

    # Returns
    - rendered image as numpy array (height, width, RGBA)
    - (config dictionary. Only if return_config was set to True.)
    '''

    model_kwargs = {'use_gpu': False}
    if 'gpu' in kwargs:
        model_kwargs['use_gpu'] = kwargs['gpu']

    config = None
    if 'config_filename' in kwargs:
        config = hp.io.load(kwargs['config_filename'])
    if 'config' in kwargs:
        config = kwargs['config']

    view_kwargs = {}
    if 'debug' in kwargs:
        view_kwargs['debug'] = kwargs['debug']
        model_kwargs['debug'] = kwargs['debug']

    return_config = False
    if 'return_config' in kwargs:
        return_config = kwargs['return_config']

    view = view_.View(**view_kwargs)
    with model_.Model(**model_kwargs) as model:
        presenter = presenter_.Presenter(view, model)

        if image is not None:
            model.update_image(image)
            if config is not None:
                model.load(config)

        if file is not None:
            model.filename = file

            ## Model loading has to wait for image loading io to finish
            def img_onload(event):
                if event.action == 'propertyChanged' and event.propertyName == 'render':
                    if config is not None:
                        model.load(config)
                    view.after(1, lambda: model.detach(img_onload))
            model.attach(img_onload)

        presenter.mainloop()
        result = np.array(model.render)
        config = model.save()

    if return_config:
        return result, config
    else:
        return result


def render(file=None, image=None, **kwargs):
    '''
    Renders given file or given image using given config.

    # Arguments:
    - config_filename: Filename of the config to apply.
    - config: Dictionary with config to apply.
    - file: Optional. Filename to open
    - image: Optional. Alternatively, an opened image can be passed as an array
        of shape (height, width, n_channels)

    # Additional optional kwargs:
    - gpu: bool. If True, PyTorch+GPU based rendering will be used (if
        installed). If False (default), defaults to NumPy+CPU rendering
    - return_config: bool. If True, returns also the config dict. False by default.

    # Returns
    - rendered image as numpy array (height, width, RGBA)
    '''

    model_kwargs = {'use_gpu': False, 'drop_tasks': False}
    if 'gpu' in kwargs:
        model_kwargs['use_gpu'] = kwargs['gpu']

    config = None
    if 'config_filename' in kwargs:
        config = hp.io.load(kwargs['config_filename'])
    if 'config' in kwargs:
        config = kwargs['config']

    view_kwargs = {}
    if 'debug' in kwargs:
        view_kwargs['debug'] = kwargs['debug']
        model_kwargs['debug'] = kwargs['debug']

    with model_.Model(**model_kwargs) as model:
        if image is not None:
            model.update_image(image)
            if config is not None:
                model.load(config)

        if file is not None:
            model.filename = file

            ## Model loading has to wait for image loading io to finish
            def img_onload(event):
                if event.action == 'propertyChanged' and event.propertyName == 'render':
                    if config is not None:
                        model.load(config)
            model.attach(img_onload)

    result = np.array(model.render)

    return result



# ------------------------------------------------------------------------------
#  File: filter.py
#  Author: Jan Kukacka
#  Date: 2/2022
# ------------------------------------------------------------------------------
#  Base class for image processing filters
# ------------------------------------------------------------------------------

class Filter(object):
    '''
    Base class for filters.
    '''

    def __init__(self):
        self.active = True
        self.cache = None

    def __call__(self, img):
        '''
        Base class call handler. Child classes need to take care of case when
        this returns None.
        '''
        if not self.active:
            return img
        if self.cache is not None:
            return self.cache

    def serialize(self):
        return {'name': self.name,
                'params': {'active': self.active}}

    @staticmethod
    def deserialize(serialization):
        raise NotImplementedError()

    def _deserialize_parent(self, serialization):
        self.active = serialization['active']



# ------------------------------------------------------------------------------
#  File: limiter.py
#  Author: Jan Kukacka
#  Date: 3/2019
# ------------------------------------------------------------------------------
#  Widget for scaler with limits and step size
#  Code from https://stackoverflow.com/a/54318377/2042751
# ------------------------------------------------------------------------------


import tkinter as tk
import tkinter.ttk as ttk

class Limiter(ttk.Scale):
    """ ttk.Scale sublass that limits the precision of values. """

    def __init__(self, *args, **kwargs):
        self.resolution = kwargs.pop('resolution', 1)  # Remove non-std kwarg.
        self.chain = kwargs.pop('command', lambda *a: None)  # Save if present.
        super(Limiter, self).__init__(*args, command=self._value_changed, **kwargs)

    def _value_changed(self, newvalue):
        # newvalue = round(float(newvalue), self.precision)
        newvalue = float(newvalue) - self.cget('from')
        newvalue = self.resolution*(newvalue // self.resolution) + self.cget('from')
        winfo_toplevel = self.winfo_toplevel()
        if not isinstance(winfo_toplevel, tk.Tk):
            winfo_toplevel = winfo_toplevel.master
        winfo_toplevel.globalsetvar(self.cget('variable'), (newvalue))
        self.chain(newvalue)  # Call user specified function.



import sys
import argparse

try:
    from importlib import metadata
except ImportError: # for Python<3.8
    import importlib_metadata as metadata
__version__ = metadata.version('image-viewer-mk2')

# argument parsing
parser = argparse.ArgumentParser(description='Image Viewer MKII: Spectral image viewer')
parser.add_argument('-i', '--input', type=str, help='Filename of the image to open', default=argparse.SUPPRESS)
parser.add_argument('-c', '--config', type=str, help='Filename of the config file to import', default=argparse.SUPPRESS)

parser.add_argument('-d', '--debug', help='Print errors to console.', action='store_true', default=argparse.SUPPRESS)
parser.add_argument('-g', '--gpu', help='Use GPU rendering (default)', action='store_true', default=argparse.SUPPRESS)
parser.add_argument('-ng', '--no_gpu', help='Use CPU rendering', action='store_true', default=argparse.SUPPRESS)


def main(args=None):
    from . import app
    kwargs = vars(parser.parse_args(sys.argv[1:]))
    kwargs2 = {}

    if 'no_gpu' in kwargs:
        kwargs2['gpu'] = not kwargs['no_gpu']
    if 'gpu' in kwargs:
        kwargs2['gpu'] = kwargs['gpu']
    if 'input' in kwargs:
        kwargs2['file'] = kwargs['input']
    if 'config' in kwargs:
        kwargs2['config_filename'] = kwargs['config']
    if 'debug' in kwargs:
        kwargs2['debug'] = kwargs['debug']

    print(f'Image Viewer MKII (v{__version__}) Jan Kukacka, 2021.')
    app.start(**kwargs2)


if __name__ == "__main__":
    try:
        main()
        sys.exit(0)
    except Exception:
        sys.exit(1)



# ------------------------------------------------------------------------------
#  File: unsharp_mask.py
#  Author: Jan Kukacka
#  Date: 2/2022
# ------------------------------------------------------------------------------
#  Implementation of Unsharp mask filter
# ------------------------------------------------------------------------------

from scipy.ndimage import gaussian_filter
import happy as hp
import numpy as np

try:
    from . import filter
except ImportError:
    from filters import filter

class UnsharpMask(filter.Filter):
    '''
    Performs unsharp masking
    '''

    name = 'unsharp_mask'

    def __init__(self, strength=1, kernel_size=1):
        '''
        # Arguments:
            - strength: Amount of sharpening / blurring
            - kernel_size: Sigma of the blurring kernel
        '''
        super().__init__()
        self.strength = strength
        self.kernel_size = kernel_size

    def __call__(self, img):
        '''
        # Arguments:
            - img: tensor with the image of shape (channels, height, width)
        # Returns:
            - img: image of the same size as the input img
        '''
        result = super().__call__(img)
        if result is not None:
            return result
        else:
            self.cache = self.call(img, self.strength, self.kernel_size)
            return self.cache

    @staticmethod
    def call(img, strength, kernel_size, **kwargs):
        '''
        Taken from development version of scikit-image
        https://github.com/scikit-image/scikit-image/blob/master/skimage/filters/_unsharp_mask.py#L20
        '''
        blurred = gaussian_filter(img, sigma=kernel_size, mode='reflect')
        result = img + (img - blurred) * strength
        return np.clip(result, 0, 1)


    def serialize(self):
        base_dict = super().serialize()
        base_dict['params']['strength'] = self.strength
        base_dict['params']['kernel_size'] = self.kernel_size
        return base_dict

    @staticmethod
    def deserialize(serialization):
        strength = serialization['strength']
        kernel_size = serialization['kernel_size']
        obj = UnsharpMask(strength, kernel_size)
        obj._deserialize_parent(serialization)
        return obj



# ------------------------------------------------------------------------------
#  File: slider.py
#  Author: Jan Kukacka
#  Date: 2/2022
# ------------------------------------------------------------------------------
#  Widget containing configuration with a slider
# ------------------------------------------------------------------------------


import tkinter as tk
import tkinter.ttk as ttk

try:
    from .limiter import Limiter
except ImportError:
    from tk_widgets.limiter import Limiter

class Slider(tk.Frame):
    def __init__(self, parent, title, variable, limit_low, limit_high, resolution, skin, *args, **kwargs):
        super().__init__(parent, bg=skin.bg_color, *args, **kwargs)

        self.skin = skin

        ## Top label
        if title is not None:
            label = tk.Label(self, text=title, fg=self.skin.fg_color,
                             bg=self.skin.bg_color, anchor=tk.NW)
            label.pack(side=tk.TOP, expand=False, fill=tk.X)

        ## Pad-x compensates for entry starting too left
        grid = tk.Frame(self, bg=self.skin.bg_color)
        grid.pack(side=tk.TOP, expand=True, fill=tk.X, padx=(3,0))

        label = ttk.Entry(grid, textvariable=variable, width=6)
        label.pack(side=tk.LEFT, expand=False, fill=tk.BOTH)
        ## Left label
        label = tk.Label(grid, text='{}'.format(limit_low),
                         fg=self.skin.fg_color, bg=self.skin.bg_color, anchor=tk.NE, width=4)
        label.pack(side=tk.LEFT, expand=False, fill=tk.BOTH)
        scale = Limiter(grid, from_=limit_low, to=limit_high, resolution=resolution,
                        orient=tk.HORIZONTAL, variable=variable)
        scale.pack(side=tk.LEFT, expand=True, fill=tk.X)
        ## Right label
        label = tk.Label(grid, text='{}'.format(limit_high),
                         fg=self.skin.fg_color, bg=self.skin.bg_color, anchor=tk.NE, width=4)
        label.pack(side=tk.LEFT, expand=False, fill=tk.BOTH)



# ------------------------------------------------------------------------------
#  File: observableset.py
#  Author: Remi Emonet
#  Date: 8/2019
#  Source: https://github.com/twitwi/vuejs-python/tree/master/vuejspython/observablecollections
#  License: MIT
# ------------------------------------------------------------------------------

from .observable import Observable

class ObservableSet(Observable, set):

	def __init__(self, *args, **kwargs):
		Observable.__init__(self)
		set.__init__(self, kwargs)

	def add(self, item):
		if item not in self:
			set.add(self, item)
			self.raiseEvent('itemsAdded', items=[item])

	def remove(self, item):
		set.remove(self, item)
		self.raiseEvent('itemsRemoved', items=[item])

	def discard(self, item):
		if item not in self:
			set.discard(self, item)
			self.raiseEvent('itemsRemoved', items=[item])

	def pop(self):
		item = set.pop(self)
		self.raiseEvent('itemsRemoved', items=[item])

	def __isub__(self, other):
		self.difference_update(self, other)
		return self

	def __ixor__(self, other):
		self.symmetric_difference_update(other)
		return self

	def __iand__(self, other):
		self.intersection_update(other)
		return self

	def __ior__(self, other):
		self.update(other)
		return self

	def difference_update(self, *args):
		for arg in args:
			difference = self.difference(arg)

		self._update(difference)

	def symmetric_difference_update(self, other):
		symmetric_difference = self.symmetric_difference(other)
		self._update(symmetric_difference)

	def intersection_update(self, *args):
		for arg in args:
			newSet = set.intersection(self, arg)

		self._update(newSet)

	def update(self, *args):
		for arg in args:
			newSet = set.union(self, arg)

		self._update(newSet)

	def clear(self):
		if (len(self) > 0):
			removedItems = []

			for item in self:
				removedItems.append(item)

			set.clear(self)
			self.raiseEvent('itemsRemoved', items=[removedItems])

	def _update(self, set2):
		elemsAdded = []
		elemsRemoved = []

		added = set2 - self
		for elem in added:
			elemsAdded.append(elem)
			set.add(self, elem)

		removed = self - set2
		for elem in removed:
			elemsRemoved.append(elem)
			set.remove(self, elem)

		if(len(elemsAdded) > 0):
			self.raiseEvent('itemsAdded', items=elemsAdded)

		if (len(elemsRemoved) > 0):
			self.raiseEvent('itemsRemoved', items=elemsRemoved)



# ------------------------------------------------------------------------------
#  File: link_button.py
#  Author: Jan Kukacka
#  Date: 2/2022
# ------------------------------------------------------------------------------
#  Link button widget
# ------------------------------------------------------------------------------

import tkinter as tk
import tkinter.ttk as ttk

class LinkButton(tk.Label):
    def __init__(self, parent, skin, *args, command=None, **kwargs):
        super().__init__(parent, *args, bg=skin.bg_color, fg=skin.fg_color, **kwargs)
        self.skin = skin
        self.default_font = tk.font.nametofont(self.cget('font'))
        self.highlight_font = self.default_font.copy()
        self.highlight_font.config(underline=1)
        self.bind('<Enter>', self.on_mouseover)
        self.bind('<Leave>', self.on_mouseout)

        self._command = None
        self._command_binding = None
        self.command = command

    @property
    def command(self):
        return self._command

    @command.setter
    def command(self, val):
        if val is not self._command:
            self._command = val
            if self._command_binding is not None:
                self.unbind('<ButtonRelease-1>', self._command_binding)
            if self._command is not None:
                self._command_binding = self.bind('<ButtonRelease-1>', self._command)

    def on_mouseover(self, event):
        self.config(fg=self.skin.fg_warning_color, font=self.highlight_font)

    def on_mouseout(self, event):
        self.config(fg=self.skin.fg_color, font=self.default_font)



# ------------------------------------------------------------------------------
#  File: toggle_button.py
#  Author: Jan Kukacka
#  Date: 2/2022
# ------------------------------------------------------------------------------
#  Toggle button widget
# ------------------------------------------------------------------------------

import tkinter as tk
import tkinter.ttk as ttk

class ToggleButton(ttk.Button):
    def __init__(self, parent, image_on, image_off, variable=None, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)

        self.image_on = tk.PhotoImage(file=str(image_on))
        self.image_off = tk.PhotoImage(file=str(image_off))

        self._variable = None
        self._variable_trace_id = None

        if variable is not None:
            self.variable = variable
        else:
            self.variable = tk.BooleanVar(value=False)


        self.bind('<Button-1>', self.on_click)

    @property
    def variable(self):
        return self._variable

    @variable.setter
    def variable(self, val):
        if val is not self._variable:
            if self._variable_trace_id is not None:
                self._variable.trace_vdelete('w', self._variable_trace_id)
            self._variable = val
            self._variable_trace_id = self._variable.trace('w', self.on_modechanged)
            self.on_modechanged()

    def on_click(self, event):
        self.variable.set(not self.variable.get())

    def on_modechanged(self, *args):
        ## Update image
        if self.variable.get():
            self.configure(image=self.image_on)
        else:
            self.configure(image=self.image_off)



# ------------------------------------------------------------------------------
#  File: skin.py
#  Author: Jan Kukacka
#  Date: 4/2020
# ------------------------------------------------------------------------------
#  Appearance definitions
# ------------------------------------------------------------------------------
import tkinter.font as tkFont
import tkinter.ttk as ttk

class Skin(object):
    '''
    Skin class with definitions of appearance properties
    '''

    def __init__(self):
        '''
        Initialize properties of a default skin
        '''

        self.bg_color = '#444444'
        # self.bg_color = '#ffffff'
        self.bg_highlight_color = '#666666'
        self.fg_color = '#FFFFFF'
        # self.fg_color = '#000000'
        self.fg_warning_color = '#E63F06'

        self.font_heading = tkFont.Font(family='ansi', size=9, weight=tkFont.BOLD)

        ttk.Style().configure("TLabelframe", background=self.bg_color)
        ttk.Style().configure("TLabelframe.Label", font=self.font_heading, background=self.bg_color, foreground=self.fg_color)
        ttk.Style().configure("TRadiobutton", background=self.bg_color, foreground=self.fg_color)
        ttk.Style().configure("TScale", background=self.bg_color, foreground=self.fg_color)
        ttk.Style().configure("TCheckbutton", background=self.bg_color, foreground=self.fg_color)
        ttk.Style().configure("Highlight.TCheckbutton", background=self.bg_highlight_color)



# ------------------------------------------------------------------------------
#  File: sortable_accordion.py
#  Author: Jan Kukacka
#  Date: 2/2022
# ------------------------------------------------------------------------------
#  TkInter implementation of a widget that supports sorting and folding of items
# ------------------------------------------------------------------------------

import tkinter as tk
import tkinter.ttk as ttk
from pathlib import Path
import os

try:
    from ..ObservableCollections.observablelist import ObservableList
    from ..ObservableCollections.observabledict import ObservableDict
    from ..ObservableCollections.observable import Observable
    from ..ObservableCollections.event import Event
    from .toggle_button import ToggleButton
    from ..utils import event_handler
    from ..utils.tk_lazy_var import BooleanVar
except ImportError:
    from ObservableCollections.observablelist import ObservableList
    from ObservableCollections.observabledict import ObservableDict
    from ObservableCollections.observable import Observable
    from ObservableCollections.event import Event
    from tk_widgets.toggle_button import ToggleButton
    from utils import event_handler
    from utils.tk_lazy_var import BooleanVar

class SortableAccordion(tk.Frame):
    '''
    '''

    def __init__(self, parent, skin, *args, **kwargs):
        super().__init__(parent, background=skin.bg_color, *args, **kwargs)
        self.skin = skin
        self.items = ObservableList()
        self.widgets = []
        self.is_repacking = False
        self.placeholder = None
        self.pack_params = dict(side=tk.TOP, fill=tk.BOTH, expand=True)

    def __del__(self):
        ## Any cleanup needed?
        super().__del__()

    def add(self, item, index=None):
        if index is None:
            index = len(self.items)
        handler = event_handler.TkVarEventHandler(self.item_onchange, item_trigger=item)
        item.sa_folded = item.folded.trace('w', handler)

        self.items.insert(index, item)
        self.widgets.insert(index, item)
        item.handle.sa_mousedown = item.handle.bind('<Button-1>', self.item_ondragstart)
        item.handle.sa_mousedrag = item.handle.bind('<B1-Motion>', self.item_ondragmove)
        self._repack()

    def remove_at(self, index):
        item = self.items[index]
        item.handle.unbind('<Button-1>', item.handle.sa_mousedown)
        item.handle.unbind('<B1-Motion>', item.handle.sa_mousedrag)
        item.folded.trace_vdelete('w', item.sa_folded)

        item.destroy()
        del self.items[index]
        del self.widgets[index]
        self._repack()

    def remove(self, item):
        self.remove_at(self.items.index(item))

    def _repack(self):
        self.is_repacking = True

        for widget in self.widgets:
            widget.pack_forget()

        for widget in self.widgets:
            widget.pack(**self.pack_params)

        self.is_repacking = False

    def item_ondragstart(self, event):
        handle = event.widget
        item = handle.item_widget

        item.config(highlightthickness=1, highlightbackground=self.skin.bg_highlight_color)

        self.placeholder = tk.Frame(self, height=item.winfo_height(), width=item.winfo_width(), background='#555555')

        self.dragged_index = self.widgets.index(item)
        self.widgets[self.dragged_index].pack_forget()
        self.widgets[self.dragged_index] = self.placeholder
        self._repack()

        handle.sa_mouseup = handle.bind('<ButtonRelease-1>', self.item_ondragstop)

        item._drag_start_x = event.x
        item._drag_start_y = event.y

        x = item.winfo_x()
        y = item.winfo_y()
        item.place(x=x, y=y, height=item.winfo_height(), width=item.winfo_width())
        item.tkraise()


    def item_ondragmove(self, event):
        if self.is_repacking:
            return

        handle = event.widget
        item = handle.item_widget
        y = item.winfo_y() - item._drag_start_y + event.y

        placeholder_index = self.widgets.index(self.placeholder)
        for i,widget in enumerate(self.widgets):
            wy = widget.winfo_y()
            if y <= wy:
                if i != placeholder_index:
                    self.widgets[placeholder_index] = self.widgets[i]
                    self.widgets[i] = self.placeholder
                    self._repack()
                break
        x = item.winfo_x() - item._drag_start_x + event.x
        y = item.winfo_y() - item._drag_start_y + event.y
        item.place(x=x, y=y, height=item.winfo_height(), width=item.winfo_width())


    def item_ondragstop(self, event):
        handle = event.widget
        item = handle.item_widget

        item.config(highlightthickness=0)

        handle.unbind('<ButtonRelease-1>', handle.sa_mouseup)

        placeholder_index = self.widgets.index(self.placeholder)
        self.widgets[placeholder_index].destroy()
        self.widgets[placeholder_index] = self.items[self.dragged_index]
        self._repack()

        self.items.sort(key=lambda item:self.widgets.index(item))

    def item_onchange(self, item_trigger):
        if item_trigger.folded.get() == False:
            for item in self.items:
                if item != item_trigger:
                    item.folded.set(True)



class AccordionItem(tk.Frame):
    '''
    '''

    def __init__(self, parent, title, skin, *args, index=None, **kwargs):
        '''
        # Arguments:
            - index: position to add the item into (in the parent accordion)
        '''
        super().__init__(parent, bg=skin.bg_color, *args, **kwargs)

        self.skin = skin

        self.header = tk.Frame(self, bg=self.skin.bg_color)
        self.header.pack(side=tk.TOP, fill=tk.X, expand=True)

        drag_image = Path(os.path.dirname(os.path.abspath(__file__))).parent/'resources'/'drag16.png'
        self.drag_image = tk.PhotoImage(file=str(drag_image))
        self.handle = tk.Label(self.header, image=self.drag_image, background=self.skin.bg_color)
        self.handle.pack(side=tk.LEFT)
        self.handle.item_widget = self

        self.title = tk.Label(self.header, text=title, bg=self.skin.bg_color,
                              fg=self.skin.fg_color,anchor=tk.W,
                              font=self.skin.font_heading)
        self.title.pack(side=tk.LEFT)

        self.folded = BooleanVar(value=True)
        self.folded.trace('w', event_handler.TkVarEventHandler(self.fold))
        image_on = Path(os.path.dirname(os.path.abspath(__file__))).parent/'resources'/'arrow_down16.png'
        image_off = Path(os.path.dirname(os.path.abspath(__file__))).parent/'resources'/'arrow_left16.png'
        self.button_fold = ToggleButton(self.header, image_on=image_on, image_off=image_off, variable=self.folded)
        self.button_fold.pack(side=tk.RIGHT, pady=1)

        self.content = tk.Frame(self, bg=self.skin.bg_color)

        parent.add(self, index)

    def fold(self):
        if self.folded.get():
            self.content.pack_forget()
        else:
            self.content.pack(side=tk.TOP, fill=tk.BOTH, expand=True)



# ------------------------------------------------------------------------------
#  File: sigmoid_norm.py
#  Author: Jan Kukacka
#  Date: 2/2022
# ------------------------------------------------------------------------------
#  Implementation of Sigmoid normalization
# ------------------------------------------------------------------------------

import numpy as np

try:
    from . import filter
except ImportError:
    from filters import filter

class SigmoidNorm(filter.Filter):
    '''
    Performs sigmoid normalization of the given image, by scaling the data
    between given percentiles to range [-1;1], which then sigmoid normalizes to
    "nearly 0" and "nearly 1". Values outside given percentile are squeezed to
    the remaining range.
    '''

    name = 'sigmoid_norm'

    def __init__(self, lower=0, upper=100, new_lower=49, new_upper=51):
        '''
        # Arguments:
            - lower: percentile between 0 and 100, must be lower than `upper`.
            - upper: percentile between 0 and 100, must be higher than `lower`.
            - new_lower: percentile between 0 and 100, where lower will be mapped on
                the sigmoid curve
            - new_upper: percentile between 0 and 100, where upper will be mapped on
                the sigmoid curve
        '''
        super().__init__()
        self.lower = lower
        self.upper = upper
        self.new_lower = new_lower
        self.new_upper = new_upper

    def __call__(self, img):
        '''
        # Arguments:
            - img: array of data to be normalized

        # Returns:
            - normalized array.
        '''
        result = super().__call__(img)
        if result is not None:
            return result
        else:
            self.cache = self.call(img, self.lower, self.upper, self.new_lower, self.new_upper)
            return self.cache


    @staticmethod
    def call(img, lower, upper, new_lower, new_upper, **kwargs):
        img_min, img_max = img.min(), img.max()
        eps = 1e-8

        low = lower/100
        high = upper/100

        lower = new_lower/100
        upper = new_upper/100
        new_low = np.log(eps + lower/(1-lower))  # eps to avoid log(0)
        new_high = np.log(upper/(1-upper+eps))   # eps to avoid division by 0
        norm_img = (new_high-new_low) * (img-low)/(high-low+eps) + new_low
        norm_img = 1/(1+np.exp(-norm_img))

        norm_min, norm_max = norm_img.min(), norm_img.max()
        scale = (img_max-img_min) / (norm_max-norm_min)
        return img_min + (norm_img-norm_min) * scale


    def serialize(self):
        base_dict = super().serialize()
        base_dict['params']['lower'] = self.lower
        base_dict['params']['upper'] = self.upper
        base_dict['params']['new_lower'] = self.new_lower
        base_dict['params']['new_upper'] = self.new_upper
        return base_dict

    @staticmethod
    def deserialize(serialization):
        lower = serialization['lower']
        upper = serialization['upper']
        new_lower = serialization['new_lower']
        new_upper = serialization['new_upper']
        obj = SigmoidNorm(lower, upper, new_lower, new_upper)
        obj._deserialize_parent(serialization)
        return obj



# ------------------------------------------------------------------------------
#  File: local_norm.py
#  Author: Jan Kukacka
#  Date: 2/2022
# ------------------------------------------------------------------------------
#  Implementation of Local contrast normalizatio
# ------------------------------------------------------------------------------

from scipy.ndimage import gaussian_filter
import happy as hp
import numpy as np

try:
    from . import filter
except ImportError:
    from filters import filter

class LocalNorm(filter.Filter):
    '''
    Performs local normalization of a given image relative to the neighborhood
    of size (kernel_size, kernel_size)  using a global cutoffself.
    '''

    name = 'local_norm'

    def __init__(self, cutoff_percentile=80, kernel_size=10):
        '''
        # Arguments:
            - kernel_size: sigma of the gaussian kernel.
            - cutoff_percentile: int between 0 and 100. Global percentile cut-off,
                preventing over-amplification of noise.
        '''
        super().__init__()
        self.cutoff_percentile = cutoff_percentile
        self.kernel_size = kernel_size

    def __call__(self, img):
        '''
        # Arguments:
            - img: tensor with the image of shape (channels, height, width)
        # Returns:
            - norm_img: image of the same size as the input img, with values
                locally normalized.
        '''
        result = super().__call__(img)
        if result is not None:
            return result
        else:
            self.cache = self.call(img, self.kernel_size, self.cutoff_percentile)
            return self.cache

    @staticmethod
    def call(img, kernel_size, cutoff_percentile, **kwargs):
        ## Compute input range
        img_min, img_max = img.min(), img.max()

        norm = gaussian_filter(img, kernel_size)
        cutoff = np.max(norm) * np.power(cutoff_percentile/100, 3)
        norm_img = img / np.maximum(norm, cutoff)
        norm_img = np.nan_to_num(norm_img)

        ## Ensure norm_img has same scale as input to enable averaging
        norm_min, norm_max = norm_img.min(), norm_img.max()
        scale = (img_max-img_min)/(norm_max-norm_min)
        norm_img = img_min + (norm_img-norm_min)*scale
        result = img+norm_img

        ## Ensure output has same range as input
        res_min, res_max = result.min(), result.max()
        scale = (img_max-img_min)/(res_max-res_min)
        result = img_min + (result-res_min)*scale

        return result


    def serialize(self):
        base_dict = super().serialize()
        base_dict['params']['cutoff_percentile'] = self.cutoff_percentile
        base_dict['params']['kernel_size'] = self.kernel_size
        return base_dict

    @staticmethod
    def deserialize(serialization):
        cutoff_percentile = serialization['cutoff_percentile']
        kernel_size = serialization['kernel_size']
        obj = LocalNorm(cutoff_percentile, kernel_size)
        obj._deserialize_parent(serialization)
        return obj



# ------------------------------------------------------------------------------
#  File: model.py
#  Author: Jan Kukacka
#  Date: 3/2021
# ------------------------------------------------------------------------------
#  Application data model component
# ------------------------------------------------------------------------------

import numpy as np
import traceback
import happy as hp
from matplotlib.colors import PowerNorm, to_hex
from multiprocessing import Process, Queue
from queue import Empty
from time import time

from PIL import ImageTk, Image

try:
    from .ObservableCollections.observablelist import ObservableList
    from .ObservableCollections.observabledict import ObservableDict
    from .ObservableCollections.observable import Observable
    from .ObservableCollections.event import Event
    from .ObservableCollections.utils import make_observable, make_plain
    from .renderer import render
    from .filters.pipeline import Pipeline
    from .filters.filter_factory import get_filter_by_name
    from .filters.local_norm import LocalNorm
    from .filters.sigmoid_norm import SigmoidNorm
    from .utils import event_handler
except ImportError:
    from ObservableCollections.observablelist import ObservableList
    from ObservableCollections.observabledict import ObservableDict
    from ObservableCollections.observable import Observable
    from ObservableCollections.event import Event
    from ObservableCollections.utils import make_observable, make_plain
    from renderer import render
    from filters.pipeline import Pipeline
    from filters.filter_factory import get_filter_by_name
    from filters.local_norm import LocalNorm
    from filters.sigmoid_norm import SigmoidNorm
    from utils import event_handler

class Model(Observable):
    '''
    Data model object
    '''

    def __init__(self, use_gpu=True, debug=False, drop_tasks=True):
        super().__init__()

        ## Setup image rendering process
        self.rendering_queue = Queue()
        self.rendered_queue = Queue()
        self.rendering_process = Process(target=render, args=(self.rendering_queue, self.rendered_queue, use_gpu, debug, drop_tasks))

        ## Setup IO process
        self.io_task_queue = Queue()
        self.io_response_queue = Queue()
        self.io_process = Process(target=reader, args=(self.io_task_queue, self.io_response_queue))
        self.n_io_pending = 0

        self._filename = None
        self._image = None
        self._color_space = 'RGB'
        self._render = None
        self.suspend_render = False

        self.response_images = None

        self.channel_props = ObservableList()

        @event_handler.requires('event')
        def on_channelprops_changed(event, obj):
            obj.raiseEvent(name='propertyChanged', propertyName='channel_props', child=event)
        self.channel_props.attach(event_handler.ObservableEventHandler(on_channelprops_changed, obj=self))
        self.channel_prop_clipboard = None


    def __enter__(self):
        self.rendering_process.start()
        self.io_process.start()
        return self

    def __exit__(self, type, value, traceback):
        ## Send termination signals
        self.rendering_queue.put(None)
        self.io_task_queue.put(None)

        ## Empty the result queues
        render = 1
        while render is not None:
            render = self.rendered_queue.get()
            if render is not None:
                self.render = render[0]

        io_response = 1
        while io_response is not None:
            io_response = self.io_response_queue.get()

        ## Join processes
        self.rendering_process.join()
        self.io_process.join()

    @property
    def filename(self):
        return self._filename

    @filename.setter
    def filename(self, val):
        if val is not self._filename:
            self._filename = val
            self.raiseEvent('propertyChanged', propertyName='filename')
            self.load_image()
            # image =
            # self.update_image(image)

    @property
    def image(self):
        return self._image

    @image.setter
    def image(self, val):
        if val is not self._image:
            self._image = val
            self.raiseEvent('propertyChanged', propertyName='image')
            e = Event('propertyChanged', self)
            e.propertyName = 'image'
            self.update_render(e)

    @property
    def color_space(self):
        return self._color_space

    @color_space.setter
    def color_space(self, val):
        if val is not self._color_space:
            self._color_space = val
            self.raiseEvent('propertyChanged', propertyName='color_space')
            self.update_cmap()

    @property
    def render(self):
        return self._render

    @render.setter
    def render(self, val):
        self._render = val
        self.raiseEvent('propertyChanged', propertyName='render')

    def load_image(self, event=None):
        task = {'type': 'load_image', 'filename':self.filename}
        self.io_task_queue.put(task)
        self.n_io_pending += 1
        self.raiseEvent('ioTask')

    def update_image(self, image):
        '''
        Used to update the image and reload channels
        '''
        self.suspend_render = True
        self.image = image
        self.update_channels()
        self.suspend_render = False
        e = Event('propertyChanged', self)
        e.propertyName = 'image'
        self.update_render(e)

    def add_filter(self, channel_index, filter_obj=None, filter_dict=None, filter_name=None):
        '''
        Appends a filter to the rendering pipeline of the given channel.

        At least one of filter_obj, filter_dict, filter_name must be specified

        # Arguments:
            - channel_index: int. Index of the channel to append the filter to.
            - filter_obj: Filter object to append
            - filter_dict: Filter object dict to append
            - filter_name: Filter type name to append
        '''
        if filter_name is not None:
            T_filter = get_filter_by_name(filter_name)
            filter_obj = T_filter()
        if filter_obj is not None:
            filter_dict = filter_obj.serialize()
        filter_dict = make_observable(filter_dict)
        self.channel_props[channel_index]['pipeline']['filters'].append(filter_dict)
        filter_dict['params'].attach(event_handler.ObservableEventHandler(self.raiseEvent, name='propertyChanged', propertyName='channel_props'))
        filter_dict['params'].attach(event_handler.ObservableEventHandler(self.update_render))

    def remove_filter(self, channel_index, filter_index):
        '''
        Removes n-th filter from the rendering pipeline of the given channel

        # Arguments:
            - channel_index: int. Index of the channel to append the filter to.
            - filter_index: int. Index of the filter to be removed.
        '''
        del self.channel_props[channel_index]['pipeline']['filters'][filter_index]


    def update_channels(self):
        ## Clear old channel props
        while len(self.channel_props) > 0:
            del self.channel_props[0]

        n_channels = self.image.shape[2]

        for channel in range(n_channels):
            channel_property = {}
            channel_property['name'] = f'Channel {channel}'
            channel_property['pipeline'] = make_observable(Pipeline([LocalNorm(80,10), SigmoidNorm(0,100,49,51)]).serialize())
            channel_property['color'] = '#ffffff'
            channel_property['visible'] = True

            channel_property = ObservableDict(channel_property)
            channel_property.attach(event_handler.ObservableEventHandler(self.raiseEvent, name='propertyChanged', propertyName='channel_props'))
            channel_property.attach(event_handler.ObservableEventHandler(self.update_render))
            channel_property['pipeline']['filters'].attach(event_handler.ObservableEventHandler(self.update_render))
            for item in channel_property['pipeline']['filters']:
                item['params'].attach(event_handler.ObservableEventHandler(self.update_render))
            self.channel_props.append(channel_property)


    def check_for_render(self):
        try:
            # render, self.histograms, self.responses = self.rendered_queue.get_nowait()
            render, self.response_images = self.rendered_queue.get_nowait()
            self.render = render
        except Empty as e:
            pass

    def check_for_io(self):
        try:
            response = self.io_response_queue.get_nowait()
            self.n_io_pending -= 1
            if response['type'] == 'load_image':
                self.update_image(response['image'])
        except Empty as e:
            pass


    @event_handler.requires('event')
    def update_render(self, event=None):
        ## Check we have all images
        if self.image is None:
            return
        ## Check render is not suspended
        if self.suspend_render:
            return

        render_task = {}
        render_task['channel_properties'] = make_plain(self.channel_props)

        ## If image has changed, pass it to the rendering thread too
        if event is not None and event.action == 'propertyChanged' and event.propertyName == 'image':
            render_task['image'] = self.image

        self.rendering_queue.put(render_task)

    def save(self):
        model_dict = {}
        model_dict['channel_props'] = make_plain(self.channel_props)
        return model_dict

    def load(self, model_dict):
        ## Suspend rendering while loading
        self.suspend_render = True

        for i, channel_property in enumerate(model_dict['channel_props']):
            if i >= len(self.channel_props):
                break
            while len(self.channel_props[i]['pipeline']['filters']) > 0:
                self.remove_filter(i, 0)

            for key, value in channel_property.items():
                if key in self.channel_props[i]:
                    if key == 'pipeline':
                        for filter in channel_property['pipeline']['filters']:
                            self.add_filter(i, filter_dict=filter)
                    else:
                        self.channel_props[i][key] = value
                else:
                    print(f'Config key {key} could not be loaded.')

        self.suspend_render = False
        self.update_render()

    def transpose_image(self):
        self.image = self.image.transpose(1,0,2)

    def autocolor(self):
        for i, channel_prop in enumerate(self.channel_props):
            channel_prop['color'] = str(to_hex(f'C{i%10}'))

    def copy_params(self, channel_index):
        self.channel_prop_clipboard = make_plain(self.channel_props[channel_index])

    def paste_params(self, channel_index):
        if self.channel_prop_clipboard is None:
            return

        ## Suspend rendering while loading
        self.suspend_render = True

        while len(self.channel_props[channel_index]['pipeline']['filters']) > 0:
            self.remove_filter(channel_index, 0)

        for key, value in self.channel_prop_clipboard.items():
            if key not in ['name', 'color', 'visible', 'pipeline']:
                self.channel_props[channel_index][key] = value
            elif key == 'pipeline':
                for filter in self.channel_prop_clipboard['pipeline']['filters']:
                    self.add_filter(channel_index, filter_dict=filter)

        self.suspend_render = False
        self.update_render()



def reader(input_queue, output_queue):
    while True:
        task = input_queue.get()

        ## Termination signal
        if task is None:
            # print('Exiting rendering thread')
            break

        try:
            response = {'type': task['type']}
            if task['type'] == 'load_image':
                filename = task['filename']
                response['image'] = load_image_internal(filename)
                # except Exception:
                #     print('Error loading image')
                #     response['image'] = None

            output_queue.put(response)

        except Exception as e:
            track = traceback.format_exc()
            print('Error in IO Thread:')
            print(track)
            response['exception'] = True
            output_queue.put(response)



    ## Signal finish of the rendered queue before quitting - it needs to be emptied
    output_queue.put(None)


def load_image_internal(filename):
    image = hp.io.load(filename)

    ## Handle MAT files:
    try:
        keys = [k for k in image.keys() if 'rec' in k]
        if len(keys) == 1:
            image = image[key]
        else:
            print('Could not load image (ambiguous keys)')
    except Exception:
        pass

    if image.ndim == 2:
        image = image[...,None]

    return image



# ------------------------------------------------------------------------------
#  File: filter_config.py
#  Author: Jan Kukacka
#  Date: 2/2022
# ------------------------------------------------------------------------------
#  Widget for displaying filter configuration
# ------------------------------------------------------------------------------

import tkinter as tk
import tkinter.ttk as ttk

try:
    from .slider import Slider
    from .link_button import LinkButton
    from ..utils import event_handler
    from ..utils.tk_lazy_var import BooleanVar, DoubleVar, IntVar
except ImportError:
    from tk_widgets.slider import Slider
    from tk_widgets.link_button import LinkButton
    from utils import event_handler
    from utils.tk_lazy_var import BooleanVar, DoubleVar, IntVar

def get_filter_widget(filter_name):
    '''
    For a given filter name, returns the corresponding filter widget class
    '''
    if filter_name == 'local_norm':
        return LocalNormConfig
    elif filter_name == 'sigmoid_norm':
        return SigmoidNormConfig
    elif filter_name == 'unsharp_mask':
        return UnsharpMaskConfig
    elif filter_name == 'gamma_correction':
        return GammaCorrectionConfig
    elif filter_name == 'frangi':
        return FrangiConfig
    elif filter_name == 'minmax_norm':
        return MinMaxNormConfig
    elif filter_name == 'gaussian_blur':
        return GaussianBlurConfig
    elif filter_name == 'anisotropic_denoising':
        return AnisotropicDenoisingConfig

class FilterConfig(tk.Frame):
    def __init__(self, parent, skin, *args, **kwargs):
        super().__init__(parent, bg=skin.bg_color, *args, **kwargs)

        self.skin = skin
        self.vars = {'active': BooleanVar(value=True)}

        self.source_params = None

        topframe = tk.Frame(self, bg=self.skin.bg_color)
        topframe.pack(side=tk.TOP, expand=False, fill=tk.X)
        ttk.Checkbutton(topframe, text='Active', variable=self.vars['active'],
                        onvalue=True, offvalue=False).pack(side=tk.LEFT)
        self.btn_remove = LinkButton(topframe, text='Remove', skin=self.skin)
        self.btn_remove.pack(side=tk.RIGHT)

        self.config_frame = tk.Frame(self, background=self.skin.bg_color)
        self.config_frame.pack(side=tk.TOP, expand=True, fill=tk.BOTH, padx=(20,0), pady=0)

    def destroy(self):
        if self.source_params is not None:
            self.source_params.detach(self.source_params_event_handler)
        super().destroy()

    def populate(self, params):
        self.source_params = params
        for key, value in params.items():
            self.vars[key].set(value)
            self.vars[key].trace('w', event_handler.TkVarEventHandler(self.update_source, key=key))
        self.source_params_event_handler = event_handler.ObservableEventHandler(self.on_sourceupdatded)
        self.source_params.attach(self.source_params_event_handler)

    def update_source(self, key):
        self.source_params[key] = self.vars[key].get()

    @event_handler.requires('event')
    def on_sourceupdatded(self, event):
        if event.action == 'itemsUpdated':
            for key, val, _ in event.items:
                if self.vars[key].get() != val:
                    self.vars[key].set(val)

    def _setup_slider(self, title, variable, limit_low, limit_high, resolution):
        slider = Slider(self.config_frame, title, variable, limit_low,
                        limit_high, resolution, skin=self.skin)
        slider.pack(side=tk.TOP, expand=True, fill=tk.BOTH, padx=0, pady=0)



class LocalNormConfig(FilterConfig):
    name = 'local_norm'
    title = 'Local contrast normalization'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.vars['kernel_size'] = DoubleVar(value=10)
        self.vars['cutoff_percentile'] = DoubleVar(value=100)

        self._setup_slider('Radius', self.vars['kernel_size'], 1, 30, 1)
        self._setup_slider('Normalizer cut-off (% of max)', self.vars['cutoff_percentile'], 0, 100, 1)


class SigmoidNormConfig(FilterConfig):
    name = 'sigmoid_norm'
    title = 'Sigmoid norm'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.vars['lower'] = DoubleVar(value=0)
        self.vars['upper'] = DoubleVar(value=100)
        self.vars['new_lower'] = DoubleVar(value=0)
        self.vars['new_upper'] = DoubleVar(value=100)

        self._setup_slider('Lower end', self.vars['lower'], 0, 100, 0.5)
        self._setup_slider('Upper end', self.vars['upper'], 0, 100, 0.5)
        self._setup_slider('Sigmoid lower end', self.vars['new_lower'], 0, 100, 0.5)
        self._setup_slider('Sigmoid upper end', self.vars['new_upper'], 0, 100, 0.5)

class UnsharpMaskConfig(FilterConfig):
    name = 'unsharp_mask'
    title = 'Unsharp mask'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.vars['kernel_size'] = DoubleVar(value=1)
        self.vars['strength'] = DoubleVar(value=1)

        self._setup_slider('Radius', self.vars['kernel_size'], 0.1, 10, 0.1)
        self._setup_slider('Strength', self.vars['strength'], 0, 10, 0.1)


class GammaCorrectionConfig(FilterConfig):
    name = 'gamma_correction'
    title = 'Gamma correction'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.vars['gamma'] = DoubleVar(value=1)
        self._setup_slider('Gamma', self.vars['gamma'], 0.01, 4, 0.01)


class FrangiConfig(FilterConfig):
    name = 'frangi'
    title = 'Vesselness filter'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.vars['scale_min'] = DoubleVar(value=1)
        self.vars['scale_max'] = DoubleVar(value=10)
        self.vars['scale_step'] = DoubleVar(value=2)
        self.vars['alpha'] = DoubleVar(value=.5)
        self.vars['beta'] = DoubleVar(value=.5)
        self.vars['gamma'] = DoubleVar(value=15)
        self._setup_slider('Sigma min', self.vars['scale_min'], 0.1, 4, 0.1)
        self._setup_slider('Sigma max', self.vars['scale_max'], 0.1, 10, 0.1)
        self._setup_slider('Sigma step', self.vars['scale_step'], 0.1, 4, 0.1)
        ## Sliders for these variables don't seem to play role
        # self._setup_slider('Alpha', self.vars['alpha'], 0, 1, 0.1)
        # self._setup_slider('Beta', self.vars['beta'], 0, 1, 0.1)
        # self._setup_slider('Gamma', self.vars['gamma'], 0, 100, 1)

class MinMaxNormConfig(FilterConfig):
    name = 'minmax_norm'
    title = 'Min-Max norm'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.vars['in_min'] = DoubleVar(value=0)
        self.vars['in_max'] = DoubleVar(value=1)
        self.vars['out_min'] = DoubleVar(value=0)
        self.vars['out_max'] = DoubleVar(value=1)

        self._setup_slider('Clip min', self.vars['in_min'], 0, 1.01, 0.01)
        self._setup_slider('Clip max', self.vars['in_max'], 0, 1.01, 0.01)
        self._setup_slider('Output min', self.vars['out_min'], 0, 1, 0.01)
        self._setup_slider('Output max', self.vars['out_max'], 0, 1, 0.01)

class GaussianBlurConfig(FilterConfig):
    name = 'gaussian_blur'
    title = 'Gaussian blur'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.vars['sigma'] = DoubleVar(value=1)
        self._setup_slider('Sigma', self.vars['sigma'], 0.1, 4, 0.1)


class AnisotropicDenoisingConfig(FilterConfig):
    name = 'anisotropic_denoising'
    title = 'Anisotropic denoising'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.vars['step_size'] = DoubleVar(value=0.15)
        self.vars['sensitivity'] = DoubleVar(value=.1)
        self.vars['n_iter'] = IntVar(value=10)

        self._setup_slider('Conductivity', self.vars['sensitivity'], 0.01, 1, .01)
        self._setup_slider('# iterations', self.vars['n_iter'], 1, 20, 1)
        ## Step size slides is hidden
        # self._setup_slider('Step size', self.vars['step_size'], 0, .25, 0.01)



# ------------------------------------------------------------------------------
#  File: panel_pipelines.py
#  Author: Jan Kukacka
#  Date: 2/2022
# ------------------------------------------------------------------------------
#  Panel with pipeline details for each channel
#  This widget does not respect the MVP architecture and for simplicity purpose
#  handles communication with the model on its own.
# ------------------------------------------------------------------------------

import tkinter as tk
import tkinter.ttk as ttk

try:
    from . import filter_config
    from . import sortable_accordion
    from ..utils import event_handler
except ImportError:
    from tk_widgets import filter_config
    from tk_widgets import sortable_accordion
    from utils import event_handler

class PanelPipelines(object):
    def __init__(self, parent, skin, var_selected_channel):
        self.parent = parent
        self.skin = skin
        self.var_selected_channel = var_selected_channel
        self.channel_variables = []

        self.accordions = []

    def recreate_items(self, channel_props):
        while len(self.accordions) > 0:
            self.accordions[-1].destroy()
            self.accordions[-1].pp_observer_target.detach(self.accordions[-1].pp_observer)
            del self.accordions[-1]

        for channel_prop in channel_props:
            accordion = sortable_accordion.SortableAccordion(self.parent, self.skin)

            for filter in channel_prop['pipeline']['filters']:
                self.create_widget(accordion, filter)

            accordion.pp_observer = event_handler.ObservableEventHandler(self.on_sourceupdatded, accordion=accordion)
            accordion.pp_observer_target = channel_prop['pipeline']['filters']
            channel_prop['pipeline']['filters'].attach(accordion.pp_observer)

            handler = event_handler.ObservableEventHandler(self.update_source, source_list=channel_prop['pipeline']['filters'])
            accordion.items.attach(handler)
            self.accordions.append(accordion)

    def create_widget(self, accordion, filter, index=None):
        T_widget = filter_config.get_filter_widget(filter['name'])
        item = sortable_accordion.AccordionItem(accordion, T_widget.title, self.skin, index=index)
        ## Needed for sorting in case there were mutliple filters of the same type
        item.pp_source_filter = filter
        widget = T_widget(item.content, self.skin)
        widget.pack(side=tk.TOP, expand=True, fill=tk.BOTH)
        widget.populate(filter['params'])
        ## Trigger removing a filter from model
        handler = event_handler.TkEventHandler(self.filterwidget_onremove, accordion=accordion, filter=filter)
        widget.btn_remove.command = handler

    @staticmethod
    def filterwidget_onremove(accordion, filter):
        accordion.pp_observer_target.pop(accordion.pp_observer_target.index(filter))

    def on_channel_selected_change(self, channel_index):
        for accordion in self.accordions:
            accordion.pack_forget()
        self.accordions[channel_index].pack(side=tk.TOP, expand=True, fill=tk.BOTH)

    @event_handler.requires('event')
    def on_sourceupdatded(self, event, accordion):
        '''
        Handler of events on model.
        '''
        if event.action == 'itemsAdded':
            for i,item in enumerate(event.items):
                self.create_widget(accordion, item, index=event.index+i)
        elif event.action == 'itemsRemoved':
            for item in event.items:
                accordion.remove_at(event.index)
        elif event.action == 'sorted':
            pass
        elif event.action == 'itemsUpdated':
            ## NOTE: changes to source items are handled by the filter config
            ##       widgets directly
            pass

    @event_handler.requires('event')
    def update_source(self, event, source_list):
        '''
        Handler of sorting events on child accordions.
        Propagates sorting back to model.
        '''
        if event.action == 'sorted':
            acc_items = event.source
            acc_filters = [item.pp_source_filter for item in acc_items]
            source_list.sort(key=acc_filters.index)



# ------------------------------------------------------------------------------
#  File: minmax_norm.py
#  Author: Jan Kukacka
#  Date: 2/2022
# ------------------------------------------------------------------------------
#  Implementation of Min-Max normalization
# ------------------------------------------------------------------------------

from scipy.ndimage import gaussian_filter
import happy as hp
import numpy as np

try:
    from . import filter
except ImportError:
    from filters import filter

class MinMaxNorm(filter.Filter):
    '''
    Adjusts the minimum and maximum of the image
    '''

    name = 'minmax_norm'

    def __init__(self, in_min=0, in_max=1, out_min=0, out_max=1):
        '''
        # Arguments:
            - in_min, in_max: floats. Define saturation levels.
            - out_min, out_max: floats. Define output range.
        '''
        super().__init__()
        self.in_min = in_min
        self.in_max = in_max
        self.out_min = out_min
        self.out_max = out_max

    def __call__(self, img):
        '''
        # Arguments:
            - img: tensor with the image of shape (channels, height, width)
        # Returns:
            - norm_img: image of the same size as the input img, with values
                locally normalized.
        '''
        result = super().__call__(img)
        if result is not None:
            return result
        else:
            self.cache = self.call(img, self.in_min, self.in_max, self.out_min, self.out_max)
            return self.cache

    @staticmethod
    def call(img, in_min, in_max, out_min, out_max, **kwargs):
        ## Compute input range
        img = np.clip(img, a_min=in_min, a_max=in_max)


        scale = (out_max-out_min)/(in_max-in_min)
        result = out_min + (img-in_min)*scale

        return result


    def serialize(self):
        base_dict = super().serialize()
        base_dict['params']['in_min'] = self.in_min
        base_dict['params']['in_max'] = self.in_max
        base_dict['params']['out_min'] = self.out_min
        base_dict['params']['out_max'] = self.out_max
        return base_dict

    @staticmethod
    def deserialize(serialization):
        in_min = serialization['in_min']
        in_max = serialization['in_max']
        out_min = serialization['out_min']
        out_max = serialization['out_max']
        obj = MinMaxNorm(in_min, in_max, out_min, out_max)
        obj._deserialize_parent(serialization)
        return obj
